"""
This file defines the 'scale_config' class responsible for all the configuration file related
activities such as parsing the config file, writing the parameters into a config file, updating the
parameters.
"""
import configparser as cp


class scale_config:
    """
    Class that handles the SCALE-Sim configuration files.
    """
    #
    def __init__(self):
        """
        __init__ method
        """
        self.run_name = "scale_run"
        # Anand: ISSUE #2. Patch
        self.use_user_bandwidth = False

        self.array_rows = 4
        self.array_cols = 4
        self.ifmap_sz_kb = 256
        self.filter_sz_kb = 256
        self.ofmap_sz_kb = 128
        self.df = 'ws'
        self.ifmap_offset = 0
        self.filter_offset = 10000000
        self.ofmap_offset = 20000000
        self.req_buf_sz_rd = 60
        self.req_buf_sz_wr = 60
        self.topofile = ""
        self.layoutfile = ""
        self.bandwidths = []
        self.valid_conf_flag = False
        self.num_bank = 1
        self.num_port = 2

        # Layout flags with default values
        self.using_ifmap_custom_layout = False
        self.ifmap_sram_bank_bandwidth = 10
        self.ifmap_sram_bank_num = 10
        self.ifmap_sram_bank_port = 2
        self.using_filter_custom_layout = False
        self.filter_sram_bank_bandwidth = 10
        self.filter_sram_bank_num = 10
        self.filter_sram_bank_port = 2

        self.valid_df_list = ['os', 'ws', 'is']

        self.sparsity_support = False
        self.sparsity_representation = ""
        # self.sparsity_N = 4
        # self.sparsity_M = 4
        self.sparsity_optimized_mapping = False
        self.sparsity_block_size = 4
        self.sparsity_rand_seed = 40
    
    # Sarbartha: Added ramulator based DRAM trace support
        self.use_ramulator_trace = False
    #
    def read_conf_file(self, conf_file_in):
        """
        Method to read the configuration file and extract all the archietctural knobs.
        """

        me = 'scale_config.' + 'read_conf_file()'

        config = cp.ConfigParser()
        config.read(conf_file_in)

        section = 'general'
        self.run_name = config.get(section, 'run_name')

        # Anand: ISSUE #2. Patch
        section = 'run_presets'
        bw_mode_string = config.get(section, 'InterfaceBandwidth')
        if bw_mode_string == 'USER':
            self.use_user_bandwidth = True
        elif bw_mode_string == 'CALC':
            self.use_user_bandwidth = False
        else:
            message = 'ERROR: ' + me
            message += 'Use either USER or CALC in InterfaceBandwidth feild. Aborting!'
            return
        
        ramulator_on = config.get(section, 'UseRamulatorTrace')
        if ramulator_on == 'True':
            self.use_ramulator_trace = True
        else:
            self.use_ramulator_trace = False
        
        # TODO Sarbartha: Should be bw
        div_factor = 1
        
        section = 'architecture_presets'
        self.array_rows = int(config.get(section, 'ArrayHeight'))
        self.array_cols = int(config.get(section, 'ArrayWidth'))
        self.ifmap_sz_kb = int(config.get(section, 'ifmapsramszkB'))
        self.filter_sz_kb = int(config.get(section, 'filtersramszkB'))
        self.ofmap_sz_kb = int(config.get(section, 'ofmapsramszkB'))
        self.ifmap_offset = int(config.get(section, 'IfmapOffset'))
        self.filter_offset = int(config.get(section, 'FilterOffset'))
        self.ofmap_offset = int(config.get(section, 'OfmapOffset'))
        self.df = config.get(section, 'Dataflow')
        self.req_buf_sz_rd = int(config.get(section, 'ReadRequestBuffer')) // div_factor
        self.req_buf_sz_wr = int(config.get(section, 'WriteRequestBuffer')) // div_factor

        layout_section = 'layout'
        self.using_ifmap_custom_layout = config.getboolean(layout_section, 'IfmapCustomLayout')
        self.using_filter_custom_layout = config.getboolean(layout_section, 'FilterCustomLayout')
        self.ifmap_sram_bank_bandwidth = int(config.get(layout_section, 'IfmapSRAMBankBandwidth'))
        self.ifmap_sram_bank_num = int(config.get(layout_section, 'IfmapSRAMBankNum'))
        self.ifmap_sram_bank_port = int(config.get(layout_section, 'IfmapSRAMBankPort'))
        self.filter_sram_bank_bandwidth = int(config.get(layout_section, 'FilterSRAMBankBandwidth'))
        self.filter_sram_bank_num = int(config.get(layout_section, 'FilterSRAMBankNum'))
        self.filter_sram_bank_port = int(config.get(layout_section, 'FilterSRAMBankPort'))
        
        # Anand: ISSUE #2. Patch
        if self.use_user_bandwidth:
            self.bandwidths = [int(x.strip())
                               for x in config.get(section, 'Bandwidth').strip().split(',')]

        if self.df not in self.valid_df_list:
            print("WARNING: Invalid dataflow")

        if config.has_section('network_presets'):  # Read network_presets
            self.topofile = config.get(section, 'TopologyCsvLoc').split('"')[1]

        # Sparsity
        section = 'sparsity'
        if config.get(section, 'SparsitySupport').lower() == 'true':
            self.sparsity_support = True
        else:
            self.sparsity_support = False

        if self.sparsity_support:
            self.sparsity_representation = config.get(section, 'SparseRep')
            # self.sparsity_N = int(config.get(section, 'NonZeroElems'))
            # self.sparsity_M = int(config.get(section, 'BlockSize'))
            if config.get(section, 'OptimizedMapping').lower() == 'true':
                self.sparsity_optimized_mapping = True
            else:
                self.sparsity_optimized_mapping = False

            if self.sparsity_optimized_mapping:
                self.sparsity_block_size = int(config.get(section, 'BlockSize'))
                assert self.sparsity_block_size <= self.array_rows, "ERROR: Invalid block size"

            self.sparsity_rand_seed = int(config.get(section, 'RandomNumberGeneratorSeed'))

        self.valid_conf_flag = True

    #
    def update_from_list(self, conf_list):
        """
        Method to update the parameters through a configuration list.
        """
        if not len(conf_list) > 11:
            print("ERROR: scale_config.update_from_list: "
                  "Incompatible number of elements in the list")

        self.run_name = conf_list[0]
        self.array_rows = int(conf_list[1])
        self.array_cols = int(conf_list[2])
        self.ifmap_sz_kb = int(conf_list[3])
        self.filter_sz_kb = int(conf_list[4])
        self.ofmap_sz_kb = int(conf_list[5])
        self.ifmap_offset = int(conf_list[6])
        self.filter_offset = int(conf_list[7])
        self.ofmap_offset = int(conf_list[8])
        self.df = conf_list[9]
        bw_mode_string = str(conf_list[10])

        assert bw_mode_string in ['CALC', 'USER'], 'Invalid mode of operation'
        if bw_mode_string == "USER":
            assert not len(conf_list) < 12, 'The user bandwidth needs to be provided'
            self.bandwidths = conf_list[11]
            self.use_user_bandwidth = True
        elif bw_mode_string == 'CALC':
            self.use_user_bandwidth = False

        if len(conf_list) == 15:
            self.topofile = conf_list[14]

        self.valid_conf_flag = True

    #
    def write_conf_file(self, conf_file_out):
        """
        Method to generate a configuration file.
        """
        if not self.valid_conf_flag:
            print('ERROR: scale_config.write_conf_file: No valid config loaded')
            return

        config = cp.ConfigParser()

        section = 'general'
        config.add_section(section)
        config.set(section, 'run_name', str(self.run_name))

        section = 'architecture_presets'
        config.add_section(section)
        config.set(section, 'ArrayHeight', str(self.array_rows))
        config.set(section, 'ArrayWidth', str(self.array_cols))

        config.set(section, 'ifmapsramszkB', str(self.ifmap_sz_kb))
        config.set(section, 'filtersramszkB', str(self.filter_sz_kb))
        config.set(section, 'ofmapsramszkB', str(self.ofmap_sz_kb))

        config.set(section, 'IfmapOffset', str(self.ifmap_offset))
        config.set(section, 'FilterOffset', str(self.filter_offset))
        config.set(section, 'OfmapOffset', str(self.ofmap_offset))

        config.set(section, 'Dataflow', str(self.df))
        config.set(section, 'Bandwidth', ','.join([str(x) for x in self.bandwidths]))

        section = 'network_presets'
        config.add_section(section)
        topofile = '"' + self.topofile + '"'
        config.set(section, 'TopologyCsvLoc', str(topofile))

        with open(conf_file_out, 'w') as configfile:
            config.write(configfile)

    #
    def set_arr_dims(self, rows=1, cols=1):
        """
        Method to set the dimensions of the PE array, with default dimensions set to 1x1.
        """
        self.array_rows = rows
        self.array_cols = cols

    #
    def set_dataflow(self, dataflow='os'):
        """
        Method to set the dataflow for the matric multiplication with Output Stationary being the
        default dataflow.
        """
        self.df = dataflow

    #
    def set_buffer_sizes_kb(self, ifmap_size_kb=1, filter_size_kb=1, ofmap_size_kb=1):
        """
        Method to set the IFMAP, Filter and OFMAP SRAM sizes, with the defaults set to 1kB.
        """
        self.ifmap_sz_kb = ifmap_size_kb
        self.filter_sz_kb = filter_size_kb
        self.ofmap_sz_kb = ofmap_size_kb

    #
    def set_topology_file(self, topofile=''):
        """
        Method to set the topology file path.
        """
        self.topofile = topofile
    
    #
    def set_layout_file(self, layoutfile=''):
        self.layoutfile = layoutfile

    #
    def set_offsets(self,
                    ifmap_offset=0,
                    filter_offset=10000000,
                    ofmap_offset=20000000
                    ):
        """
        Method to set the offsets used for IFMAP, Filter and OFMAP addresses, with the defaults set
        to 0, 10M and 20M respectively.
        """
        self.ifmap_offset = ifmap_offset
        self.filter_offset = filter_offset
        self.ifmap_offset = ofmap_offset
        self.valid_conf_flag = True

    #
    def force_valid(self):
        """
        Method to set the 'valid_config_flag' without any checks.
        """
        self.valid_conf_flag = True

    #
    def set_bw_mode_to_calc(self):
        """
        Method to set the 'use_user_bandwidth' to CALC mode.
        """
        self.use_user_bandwidth = False

    #
    def use_user_dram_bandwidth(self):
        """
        Method that returns the value of 'use_user_bandwidth'.
        """
        if not self.valid_conf_flag:
            me = 'scale_config.' + 'use_user_dram_bandwidth()'
            message = 'ERROR: ' + me + ': Configuration is not valid'
            print(message)
            return

        return self.use_user_bandwidth

    #
    def get_conf_as_list(self):
        """
        Method to extract the configuration parameters in the form of a list.
        """
        out_list = []

        if not self.valid_conf_flag:
            print("ERROR: scale_config.get_conf_as_list: Configuration is not valid")
            return

        out_list.append(str(self.run_name))

        out_list.append(str(self.array_rows))
        out_list.append(str(self.array_cols))

        out_list.append(str(self.ifmap_sz_kb))
        out_list.append(str(self.filter_sz_kb))
        out_list.append(str(self.ofmap_sz_kb))

        out_list.append(str(self.ifmap_offset))
        out_list.append(str(self.filter_offset))
        out_list.append(str(self.ofmap_offset))

        out_list.append(str(self.df))
        out_list.append(str(self.topofile))

        return out_list

    #
    def get_run_name(self):
        """
        Method to get the run name used for the simulation.
        """
        if not self.valid_conf_flag:
            print("ERROR: scale_config.get_run_name() : Config data is not valid")
            return

        return self.run_name

    #
    def get_topology_path(self):
        """
        Method to get the topology file path used for the simulation.
        """
        if not self.valid_conf_flag:
            print("ERROR: scale_config.get_topology_path() : Config data is not valid")
            return
        return self.topofile

    def get_layout_path(self):
        if not self.valid_conf_flag:
            print("ERROR: scale_config.get_layout_path() : Config data is not valid")
            return
        return self.layoutfile

    def get_topology_name(self):
        """
        Method to extract the name of the topology file from the topology path.
        """
        if not self.valid_conf_flag:
            print("ERROR: scale_config.get_topology_name() : Config data is not valid")
            return

        name = self.topofile.split('/')[-1].strip()
        name = name.split('.')[0]

        return name

    #
    def get_dataflow(self):
        """
        Method to get the dataflow used for the simulation.
        """
        if self.valid_conf_flag:
            return self.df

    #
    def get_array_dims(self):
        """
        Method to get the dimensions of the PE array.
        """
        if self.valid_conf_flag:
            return self.array_rows, self.array_cols

    #
    def get_mem_sizes(self):
        """
        Method to get the IFMAP, Filter and OFMAP SRAM sizes.
        """
        me = 'scale_config.' + 'get_mem_sizes()'

        if not self.valid_conf_flag:
            message = 'ERROR: ' + me
            message += 'Config is not valid. Not returning any values'
            return

        return self.ifmap_sz_kb, self.filter_sz_kb, self.ofmap_sz_kb

    #
    def get_offsets(self):
        """
        Method to get the offsets used for IFMAP, Filter and OFMAP addresses.
        """
        if self.valid_conf_flag:
            return self.ifmap_offset, self.filter_offset, self.ofmap_offset
    
    def get_ramulator_trace(self):
        """
        Method to check if the run considers ramulator trace numpy files
        """
        if self.valid_conf_flag:
            return self.use_ramulator_trace
    
    def get_req_buf_sz_rd(self):
        """
        Method to set the read request buffer size
        """
        if self.valid_conf_flag:
            return self.req_buf_sz_rd
    
    def get_req_buf_sz_wr(self):
        """
        Method to set the write request buffer size
        """
        if self.valid_conf_flag:
            return self.req_buf_sz_wr
    
    #
    def get_bandwidths_as_string(self):
        """
        Method to get the bandwidths as a string.
        """
        if self.valid_conf_flag:
            return ','.join([str(x) for x in self.bandwidths])

  #
    def get_ifmap_sram_bandwidth(self):
        """
        Method to get the bandwidths as a value.
        """
        if self.valid_conf_flag:
            return self.ifmap_sram_bank_bandwidth

    def get_filter_sram_bandwidth(self):
      """
      Method to get the bandwidths as a value.
      """
      if self.valid_conf_flag:
          return self.filter_sram_bank_bandwidth     

    #
    def get_bandwidths_as_list(self):
        """
        Method to get the bandwidths as a list.
        """
        if self.valid_conf_flag:
            return self.bandwidths

    def get_bandwidths_as_list(self):
        if self.valid_conf_flag:
            return self.bandwidths
        
    def get_num_bank(self):
        if self.valid_conf_flag:
            return self.num_bank
        
    def get_num_port(self):
        if self.valid_conf_flag:
            return self.num_port
        
    def get_min_dram_bandwidth(self):
        """
        Method to get the minimum DRAM bandwidth defined in the configuration.
        """
        if not self.use_user_dram_bandwidth():
            me = 'scale_config.' + 'get_min_dram_bandwidth()'
            message = 'ERROR: ' + me + ': No user bandwidth provided'
            print(message)
        else:
            return min(self.bandwidths)

    # FIX ISSUE #14
    @staticmethod
    def get_default_conf_as_list():
        """
        Method to get the default configuration as a list.
        """
        dummy_obj = scale_config()
        dummy_obj.force_valid()
        out_list = dummy_obj.get_conf_as_list()
        return out_list



"""
Double buffer read memory implementation
"""
# TODO: Verification Pending
import math
import numpy as np
from tqdm import tqdm

from scalesim.memory.read_port import read_port


class read_buffer:
    """
    Class which runs the memory simulation of double buffered ifmap/filter SRAM. The double
    buffering helps to hide the DRAM latency when the SRAM is servicing requests from the systolic
    array using one of the buffers while the other buffer prefetches from the DRAM.
    """
    #
    def __init__(self):
        """
        __init__ method.
        """
        # Buffer properties: User specified
        self.total_size_bytes = 128
        self.word_size = 1         # Bytes
        self.active_buf_frac = 0.9
        self.hit_latency = 1       # Cycles after which a request is served if already in the buffer

        # Buffer properties: Calculated
        self.total_size_elems = math.floor(self.total_size_bytes / self.word_size)
        self.active_buf_size = int(math.ceil(self.total_size_elems * 0.9))
        self.prefetch_buf_size = self.total_size_elems - self.active_buf_size

        # Backing interface properties
        self.backing_buffer = read_port()
        self.req_gen_bandwidth = 100            # words per cycle

        # Status of the buffer
        self.hashed_buffer = {}
        self.active_buffer_set_limits = []
        self.prefetch_buffer_set_limits = []
        self.active_buffer_contents = set()

        # Variables to enable prefetching
        self.fetch_matrix = np.ones((1, 1))
        self.last_prefect_cycle = -1
        self.next_line_prefetch_idx = 0
        self.next_col_prefetch_idx = 0
        self.prefetch_buffer_contents = []

        # Access counts
        self.num_access = 0

        # Trace matrix
        self.trace_matrix = np.ones((1, 1))

        # Flags
        self.active_buf_full_flag = False
        self.hashed_buffer_valid = False
        self.trace_valid = False

    #
    def set_params(self, backing_buf_obj,
                   total_size_bytes=1, word_size=1, active_buf_frac=0.9,
                   hit_latency=1, backing_buf_bw=1
                   ):
        """
        Method to set the ifmap/filter double buffered memory simulation parameters for
        housekeeping.
        """

        self.total_size_bytes = total_size_bytes
        self.word_size = word_size

        assert 0.5 <= active_buf_frac < 1, "Valid active buf frac [0.5,1)"
        self.active_buf_frac = round(active_buf_frac, 2)
        self.hit_latency = hit_latency

        self.backing_buffer = backing_buf_obj
        self.req_gen_bandwidth = backing_buf_bw

        # Calculate these based on the values provided
        self.total_size_elems = math.floor(self.total_size_bytes / self.word_size)
        self.active_buf_size = int(math.ceil(self.total_size_elems * self.active_buf_frac))
        self.prefetch_buf_size = self.total_size_elems - self.active_buf_size

    #
    def reset(self): # TODO: check if all resets are working propoerly
        """
        Method to reset the read buffer parameters.
        """
        # Buffer properties: User specified
        self.total_size_bytes = 128
        self.word_size = 1  # Bytes
        self.active_buf_frac = 0.9
        self.hit_latency = 1  # Cycles after which a request is served if already in the buffer

        # Buffer properties: Calculated
        self.total_size_elems = math.floor(self.total_size_bytes / self.word_size)
        self.active_buf_size = int(math.ceil(self.total_size_elems * 0.9))
        self.prefetch_buf_size = self.total_size_elems - self.active_buf_size

        # Backing interface properties
        self.backing_buffer = read_port()
        self.req_gen_bandwidth = 100  # words per cycle

        # Status of the buffer
        self.hashed_buffer = {}
        self.active_buffer_set_limits = []
        self.prefetch_buffer_set_limits = []
        self.active_buffer_contents = set()

        # Variables to enable prefetching
        self.fetch_matrix = np.ones((1, 1))
        self.last_prefect_cycle = -1
        self.next_line_prefetch_idx = 0
        self.next_col_prefetch_idx = 0
        self.prefetch_buffer_contents = []

        # Access counts
        self.num_access = 0

        # Trace matrix
        self.trace_matrix = np.ones((1, 1))

        # Flags
        self.active_buf_full_flag = False
        self.hashed_buffer_valid = False
        self.trace_valid = False

    #
    def set_fetch_matrix(self, fetch_matrix_np):
        """
        Method to set the fetch matrix responsible for prefetching from the DRAM.
        """
        # The operand matrix determines what to pre-fetch into both active and prefetch buffers
        # req_gen_bandwidth is set to 100 by default and will be used in 'calc' mode
        # In 'user' mode, this will be set in the set_params

        num_elems = fetch_matrix_np.shape[0] * fetch_matrix_np.shape[1]
        num_lines = int(math.ceil(num_elems / self.req_gen_bandwidth))
        self.fetch_matrix = np.ones((num_lines, self.req_gen_bandwidth)) * -1

        # Put stuff into the fetch matrix
        # This is done to ensure that there is no shape mismatch
        # Not sure if this is the optimal way to do it or not
        for i in range(num_elems):
            src_row = math.floor(i / fetch_matrix_np.shape[1])
            src_col = math.floor(i % fetch_matrix_np.shape[1])

            dest_row = math.floor(i / self.req_gen_bandwidth)
            dest_col = math.floor(i % self.req_gen_bandwidth)

            self.fetch_matrix[dest_row][dest_col] = fetch_matrix_np[src_row][src_col]

        # Once the fetch matrices are set, populate the data structure for faster lookups and
        # servicing
        self.prepare_hashed_buffer()

    #
    def prepare_hashed_buffer(self):
        """
        Method to convert the fetch matrix into a hashed buffer for fast lookups.
        """
        elems_per_set = math.ceil(self.total_size_elems / 100)

        prefetch_rows = self.fetch_matrix.shape[0]
        prefetch_cols = self.fetch_matrix.shape[1]

        line_id = 0
        elem_ctr = 0
        current_line = set()

        for r in range(prefetch_rows):
            for c in range(prefetch_cols):
                elem = self.fetch_matrix[r][c]

                if not elem == -1:
                    current_line.add(elem)
                    elem_ctr += 1

                if not elem_ctr < elems_per_set:    # ie > or =
                    self.hashed_buffer[line_id] = current_line
                    line_id += 1
                    elem_ctr = 0
                    current_line = set()        # new set

        self.hashed_buffer[line_id] = current_line
        self.hashed_buffer_valid = True

    #
    def service_reads(self,
                      incoming_requests_arr_np, # 2D array with the requests
                      incoming_cycles_arr):     # 1D vector with the cycles at which req arrived
        """
        Method to service read requests coming from systolic array.
        """
        # Service the incoming read requests
        # returns a cycles array corresponding to the requests buffer
        # Logic: Always check if an addr is in active buffer.
        #        If hit, return with hit latency
        #        Else, make the contents of prefetch buffer as active and then check
        #              finish till an ongoing prefetch is done before reassiging prefetch buffer

        if not self.active_buf_full_flag:
            start_cycle = incoming_cycles_arr[0][0]
            # Needs to use the entire operand matrix
            # keeping in mind the tile order and everything
            self.prefetch_active_buffer(start_cycle=start_cycle)

        out_cycles_arr = []
        offset = self.hit_latency
        #for cycle, request_line in tqdm(zip(incoming_cycles_arr, incoming_requests_arr_np)):
        for i in tqdm(range(incoming_requests_arr_np.shape[0])):
            cycle = incoming_cycles_arr[i]
            request_line = set(incoming_requests_arr_np[i]) #shaves off a few seconds

            for addr in request_line:
                if addr == -1:
                    continue

                if addr not in self.active_buffer_contents: #this is super slow!!!
                    self.new_prefetch()
                    potential_stall_cycles = self.last_prefect_cycle - (cycle + offset)
                    # Offset increments if there were potential stalls
                    offset += potential_stall_cycles

            out_cycles = cycle + offset
            out_cycles_arr.append(out_cycles)

        out_cycles_arr_np = np.asarray(out_cycles_arr).reshape((len(out_cycles_arr), 1))

        return out_cycles_arr_np

    #
    def prefetch_active_buffer(self, start_cycle):
        """
        Method to prefetch the active read buffer before servicing individual memory requests.
        """
        # Depending on size of the active buffer, calculate the number of lines from op mat to fetch
        # Also, calculate the cycles arr for requests

        # 1. Preparing the requests:
        num_lines = math.ceil(self.active_buf_size / self.req_gen_bandwidth)
        if not num_lines < self.fetch_matrix.shape[0]:
            num_lines = self.fetch_matrix.shape[0]

        requested_data_size = num_lines * self.req_gen_bandwidth
        self.num_access += requested_data_size

        start_idx = 0
        end_idx = num_lines

        prefetch_requests = self.fetch_matrix[start_idx:end_idx, :]
        # 1.1 See if extra requests are made, if so nullify them
        self.next_col_prefetch_idx = 0
        if requested_data_size > self.active_buf_size:
            valid_cols = int(self.active_buf_size % self.req_gen_bandwidth)
            row = end_idx - 1
            self.next_col_prefetch_idx = valid_cols
            for col in range(valid_cols, self.req_gen_bandwidth):
                prefetch_requests[row][col] = -1

        # 2. Preparing the cycles array
        #    The start_cycle variable ensures that all the requests have been made before any
        #    incoming reads came
        cycles_arr = np.zeros((num_lines, 1))
        for i in range(cycles_arr.shape[0]):
            cycles_arr[i][0] = \
                -1 * (num_lines - start_cycle - (i - self.backing_buffer.get_latency()))

        # 3. Send the request and get the response cycles count
        response_cycles_arr = \
            self.backing_buffer.service_reads(incoming_cycles_arr=cycles_arr,
                                              incoming_requests_arr_np=prefetch_requests)

        # 4. Update the variables
        self.last_prefect_cycle = int(response_cycles_arr[-1][0])

        # Update the trace matrix
        self.trace_matrix = np.concatenate((response_cycles_arr, prefetch_requests), axis=1)
        self.trace_valid = True

        # Set active buffer contents
        num_elems_to_fetch = min(self.active_buf_size, requested_data_size)
        req_col_size = prefetch_requests.shape[1]
        for i in range(num_elems_to_fetch):
            r = int(i / req_col_size)
            c = int(i % req_col_size)

            addr = prefetch_requests[r][c]
            if not addr == -1:
                self.active_buffer_contents.add(addr)
        self.active_buf_full_flag = True

        # Set the line to be prefetched next
        if requested_data_size > self.active_buf_size:
            self.next_line_prefetch_idx = num_lines % self.fetch_matrix.shape[0]
        else:
            self.next_line_prefetch_idx = (num_lines + 1) % self.fetch_matrix.shape[0]

    #
    def new_prefetch(self):
        """
        Method to do a new prefetch. In a new prefetch, some portion of the original data needs to
        be deleted to accomodate the prefetched data In this case we overwrite some data in the
        active buffer with the prefetched data and then create a new prefetch request.
        """
        # In a new prefetch, some portion of the original data needs to be deleted to accomodate the
        # prefetched data.
        # In this case we overwrite some data in the active buffer with the prefetched data
        # And then create a new prefetch request
        # Also return when the prefetched data was made available

        # 1. Rewrite the active buffer
        assert self.active_buf_full_flag, 'Active buffer is empty'
        for i in range(len(self.prefetch_buffer_contents)):
            data = self.prefetch_buffer_contents[i]
            self.active_buffer_contents.add(data)

        self.prefetch_buffer_contents = []

        # 2. Create the request
        start_idx = self.next_line_prefetch_idx
        num_lines = math.ceil(self.prefetch_buf_size / self.req_gen_bandwidth)
        end_idx = start_idx + num_lines
        requested_data_size = num_lines * self.req_gen_bandwidth
        self.num_access += requested_data_size

        # In care we need to circle back
        if end_idx > self.fetch_matrix.shape[0]:
            last_idx = self.fetch_matrix.shape[0]
            prefetch_requests = self.fetch_matrix[start_idx:,:]

            new_end_idx = min(end_idx - last_idx, start_idx) # In case the entire array is engulfed
            prefetch_requests = \
                np.concatenate((prefetch_requests, self.fetch_matrix[:new_end_idx,:]))
        else:
            prefetch_requests = self.fetch_matrix[start_idx:end_idx, :]

        # Modify the prefetch request to drop unwanted addresses
        # a. Chomp the elements in the first line included in previous fetches
        for i in range(0, self.next_col_prefetch_idx):
            prefetch_requests[0][i] = -1

        # b. Chomp the excess elements in the last line
        if requested_data_size > self.active_buf_size:
            valid_cols = int(self.active_buf_size % self.req_gen_bandwidth)
            row = end_idx - 1
            for col in range(valid_cols, self.req_gen_bandwidth):
                prefetch_requests[row][col] = -1

        # 3. Create the request cycles
        cycles_arr = np.zeros((num_lines, 1))
        for i in range(cycles_arr.shape[0]):
            cycles_arr[i][0] = self.last_prefect_cycle + i

        # 4. Send the request
        response_cycles_arr = \
            self.backing_buffer.service_reads(incoming_cycles_arr=cycles_arr,
                                              incoming_requests_arr_np=prefetch_requests)

        # 5. Update the variables
        self.last_prefect_cycle = response_cycles_arr[-1][0]

        this_prefetch_trace = np.concatenate((response_cycles_arr, prefetch_requests), axis=1)
        self.trace_matrix = np.concatenate((self.trace_matrix, this_prefetch_trace), axis=0)

        # Set the contents of the prefetch buffer
        num_prefetch_reqs = prefetch_requests.shape[0] * prefetch_requests.shape[1]
        elems_to_prefectch = min(self.prefetch_buf_size, num_prefetch_reqs)
        req_col_size = prefetch_requests.shape[1]
        for i in range(elems_to_prefectch):
            r = int(i / req_col_size)
            c = int(i % req_col_size)

            addr = prefetch_requests[r][c]
            self.prefetch_buffer_contents.append(addr)

        # Set the line to be prefetched next
        if requested_data_size > self.active_buf_size:
            self.next_line_prefetch_idx = num_lines % self.fetch_matrix.shape[0]
        else:
            self.next_line_prefetch_idx = (num_lines + 1) % self.fetch_matrix.shape[1]

        # This does not need to return anything
        #return response_cycles_arr

    #
    def get_trace_matrix(self):
        """
        Method to get the read buffer trace matrix. It contains addresses requsted by the systolic
        array and the cycles (first column) at which the requests are made.
        """
        if not self.trace_valid:
            print('No trace has been generated yet')
            return

        return self.trace_matrix

    #
    def get_hit_latency(self):
        """
        Method to get hit latency of the read buffer.
        """
        return self.hit_latency

    #
    def get_num_accesses(self):
        """
        Method to get number of accesses of the read buffer if trace_valid flag is set.
        """
        assert self.trace_valid, 'Traces not ready yet'
        return self.num_access

    #
    def get_external_access_start_stop_cycles(self):
        """
        Method to get start and stop cycles of the read buffer if trace_valid flag is set.
        """
        assert self.trace_valid, 'Traces not ready yet'
        start_cycle = self.trace_matrix[0][0]
        end_cycle = self.trace_matrix[-1][0]

        return start_cycle, end_cycle

    #
    def print_trace(self, filename):
        """
        Method to write the read buffer trace matrix to a file.
        """
        if not self.trace_valid:
            print('No trace has been generated yet')
            return

        np.savetxt(filename, self.trace_matrix, fmt='%s', delimiter=",")



"""
This file contains the 'simulator' class that simulates the entire model using the class
'single_layer_sim' and generates the reports (.csv files).
"""

import os

from scalesim.scale_config import scale_config as cfg
from scalesim.topology_utils import topologies as topo
from scalesim.layout_utils import layouts as layout
from scalesim.single_layer_sim import single_layer_sim as layer_sim


class simulator:
    """
    Class which runs the simulations and manages generated data across various layers
    """
    #
    def __init__(self):
        """
        __init__ method
        """
        self.conf = cfg()
        self.topo = topo()
        self.layout = layout()

        self.top_path = "./"
        self.verbose = True
        self.save_trace = True

        self.num_layers = 0

        self.single_layer_sim_object_list = []

        self.params_set_flag = False
        self.all_layer_run_done = False

    #
    def set_params(self,
                   config_obj=cfg(),
                   topo_obj=topo(),
                   layout_obj=layout(),
                   top_path="./",
                   verbosity=True,
                   save_trace=True
                   ):
        """
        Method to set the run parameters including inputs and parameters for housekeeping.
        """
        self.conf = config_obj
        self.topo = topo_obj
        self.layout = layout_obj

        self.top_path = top_path
        self.verbose = verbosity
        self.save_trace = save_trace

        # Calculate inferrable parameters here
        self.num_layers = self.topo.get_num_layers()

        self.params_set_flag = True

    #
    def run(self):
        """
        Method to run scalesim simulation for all layers. This method first runs compute and memory
        simulations for each layer and gathers the required stats. Once the simulation runs are
        done, it gathers the stats from single_layer_sim objects and calls generate_report() method
        to create the report files. If save_trace flag is set, then layer wise traces are saved as
        well.
        """
        assert self.params_set_flag, 'Simulator parameters are not set'

        # 1. Create the layer runners for each layer
        for i in range(self.num_layers):
            this_layer_sim = layer_sim()
            this_layer_sim.set_params(layer_id=i,
                                 config_obj=self.conf,
                                 topology_obj=self.topo,
                                 layout_obj=self.layout,
                                 verbose=self.verbose)

            self.single_layer_sim_object_list.append(this_layer_sim)

        if not os.path.isdir(self.top_path):
            os.mkdir(self.top_path)

        report_path = self.top_path + '/' + self.conf.get_run_name()

        if not os.path.isdir(report_path):
            os.mkdir(report_path)

        self.top_path = report_path

        # 2. Run each layer
        # TODO: This is parallelizable
        for single_layer_obj in self.single_layer_sim_object_list:

            if self.verbose:
                layer_id = single_layer_obj.get_layer_id()
                print('\nRunning Layer ' + str(layer_id))

            single_layer_obj.run()

            if self.verbose:
                comp_items = single_layer_obj.get_compute_report_items()
                total_cycles = comp_items[0]
                comp_cycles = comp_items[1]
                stall_cycles = comp_items[2]
                util = comp_items[3]
                mapping_eff = comp_items[4]
                print('Total cycles: ' + str(total_cycles))
                print('Compute cycles: ' + str(comp_cycles))
                print('Stall cycles: ' + str(stall_cycles))
                print('Overall utilization: ' + "{:.2f}".format(util) +'%')
                print('Mapping efficiency: ' + "{:.2f}".format(mapping_eff) +'%')

                avg_bw_items = single_layer_obj.get_bandwidth_report_items()
                if self.conf.sparsity_support is True:
                    avg_ifmap_sram_bw = avg_bw_items[0]
                    avg_filter_sram_bw = avg_bw_items[1]
                    avg_filter_metadata_sram_bw = avg_bw_items[2]
                    avg_ofmap_sram_bw = avg_bw_items[3]
                    avg_ifmap_dram_bw = avg_bw_items[4]
                    avg_filter_dram_bw = avg_bw_items[5]
                    avg_ofmap_dram_bw = avg_bw_items[6]
                else:
                    avg_ifmap_sram_bw = avg_bw_items[0]
                    avg_filter_sram_bw = avg_bw_items[1]
                    avg_ofmap_sram_bw = avg_bw_items[2]
                    avg_ifmap_dram_bw = avg_bw_items[3]
                    avg_filter_dram_bw = avg_bw_items[4]
                    avg_ofmap_dram_bw = avg_bw_items[5]

                print('Average IFMAP SRAM BW: ' + "{:.3f}".format(avg_ifmap_sram_bw) + \
                      ' words/cycle')
                print('Average Filter SRAM BW: ' + "{:.3f}".format(avg_filter_sram_bw) + \
                      ' words/cycle')
                if self.conf.sparsity_support is True:
                    print('Average Filter Metadata SRAM BW: ' + \
                          "{:.3f}".format(avg_filter_metadata_sram_bw) + ' words/cycle')
                print('Average OFMAP SRAM BW: ' + "{:.3f}".format(avg_ofmap_sram_bw) + \
                      ' words/cycle')
                print('Average IFMAP DRAM BW: ' + "{:.3f}".format(avg_ifmap_dram_bw) + \
                      ' words/cycle')
                print('Average Filter DRAM BW: ' + "{:.3f}".format(avg_filter_dram_bw) + \
                      ' words/cycle')
                print('Average OFMAP DRAM BW: ' + "{:.3f}".format(avg_ofmap_dram_bw) + \
                      ' words/cycle')

            if self.save_trace:
                if self.verbose:
                    print('Saving traces: ', end='')
                single_layer_obj.save_traces(self.top_path)
                if self.verbose:
                    print('Done!')

        self.all_layer_run_done = True

        self.generate_reports()

    #
    def generate_reports(self):
        """
        Method to generate the report files for scalesim run if the runs are already completed. For
        each layer, this method collects the report data from single_layer_sim objects and then
        prints them out into COMPUTE_REPORT.csv, BANDWIDTH_REPORT.csv, DETAILED_ACCESS_REPORT.csv
        and SPARSE_REPORT.csv files.
        """
        assert self.all_layer_run_done, 'Layer runs are not done yet'

        compute_report_name = self.top_path + '/COMPUTE_REPORT.csv'
        compute_report = open(compute_report_name, 'w')
        header = ('LayerID, Total Cycles (incl. prefetch), Total Cycles, Stall Cycles, Overall Util %, Mapping Efficiency %,'
                  ' Compute Util %,\n')
        compute_report.write(header)

        bandwidth_report_name = self.top_path + '/BANDWIDTH_REPORT.csv'
        bandwidth_report = open(bandwidth_report_name, 'w')
        if self.conf.sparsity_support is True:
            header = ('LayerID, Avg IFMAP SRAM BW, Avg FILTER SRAM BW, Avg FILTER Metadata SRAM BW,'
                      ' Avg OFMAP SRAM BW, ')
        else:
            header = 'LayerID, Avg IFMAP SRAM BW, Avg FILTER SRAM BW, Avg OFMAP SRAM BW, '
        header += 'Avg IFMAP DRAM BW, Avg FILTER DRAM BW, Avg OFMAP DRAM BW,\n'
        bandwidth_report.write(header)

        detail_report_name = self.top_path + '/DETAILED_ACCESS_REPORT.csv'
        detail_report = open(detail_report_name, 'w')
        header = 'LayerID, '
        header += 'SRAM IFMAP Start Cycle, SRAM IFMAP Stop Cycle, SRAM IFMAP Reads, '
        header += 'SRAM Filter Start Cycle, SRAM Filter Stop Cycle, SRAM Filter Reads, '
        header += 'SRAM OFMAP Start Cycle, SRAM OFMAP Stop Cycle, SRAM OFMAP Writes, '
        header += 'DRAM IFMAP Start Cycle, DRAM IFMAP Stop Cycle, DRAM IFMAP Reads, '
        header += 'DRAM Filter Start Cycle, DRAM Filter Stop Cycle, DRAM Filter Reads, '
        header += 'DRAM OFMAP Start Cycle, DRAM OFMAP Stop Cycle, DRAM OFMAP Writes,\n'
        detail_report.write(header)

        if self.conf.sparsity_support is True:
            sparse_report_name = self.top_path + '/SPARSE_REPORT.csv'
            sparse_report = open(sparse_report_name, 'w')
            header = 'LayerID, '
            header += 'Sparsity Representation, '
            header += ('Original Filter Storage, New Storage (Filter+Metadata),'
                       ' Filter Metadata Storage, ')
            header += 'Avg FILTER Metadata SRAM BW, '
            header += '\n'
            sparse_report.write(header)

        for lid in range(len(self.single_layer_sim_object_list)):
            single_layer_obj = self.single_layer_sim_object_list[lid]
            compute_report_items_this_layer = single_layer_obj.get_compute_report_items()
            log = str(lid) +', '
            log += ', '.join([str(x) for x in compute_report_items_this_layer])
            log += ',\n'
            compute_report.write(log)

            bandwidth_report_items_this_layer = single_layer_obj.get_bandwidth_report_items()
            log = str(lid) + ', '
            log += ', '.join([str(x) for x in bandwidth_report_items_this_layer])
            log += ',\n'
            bandwidth_report.write(log)

            detail_report_items_this_layer = single_layer_obj.get_detail_report_items()
            log = str(lid) + ', '
            log += ', '.join([str(x) for x in detail_report_items_this_layer])
            log += ',\n'
            detail_report.write(log)

            if self.conf.sparsity_support is True:
                sparse_report_items_this_layer = single_layer_obj.get_sparse_report_items()
                log = str(lid) + ', ' + self.conf.sparsity_representation + ', '
                log += ', '.join([str(x) for x in sparse_report_items_this_layer])
                log += ',\n'
                sparse_report.write(log)

        compute_report.close()
        bandwidth_report.close()
        detail_report.close()
        if self.conf.sparsity_support is True:
            sparse_report.close()

    #
    def get_total_cycles(self):
        """
        Method which aggregates the total cycles (both compute and stall) across all the layers for
        the given workload.
        """
        assert self.all_layer_run_done, 'Layer runs are not done yet'

        total_cycles = 0
        for layer_obj in self.single_layer_sim_object_list:
            cycles_this_layer = int(layer_obj.get_compute_report_items[0])
            total_cycles += cycles_this_layer

        return total_cycles



"""
This file contains the 'single_layer_sim' class that handles the simulation for a single layer of
the model and generates the data that goes inside the reports.
"""

import os
import numpy as np

from scalesim.compute.compression import compression as cp

from scalesim.scale_config import scale_config as cfg
from scalesim.topology_utils import topologies as topo
from scalesim.layout_utils import layouts as layout
from scalesim.compute.operand_matrix import operand_matrix as opmat
from scalesim.compute.systolic_compute_os import systolic_compute_os
from scalesim.compute.systolic_compute_ws import systolic_compute_ws
from scalesim.compute.systolic_compute_is import systolic_compute_is
from scalesim.memory.double_buffered_scratchpad_mem import double_buffered_scratchpad as mem_dbsp

class single_layer_sim:
    """
    Class which runs the simulation for a single layer and generates report data
    """
    #
    def __init__(self):
        """
        __init__ method
        """
        self.layer_id = 0
        self.dataflow = ''
        self.topo = topo()
        self.layout = layout()
        self.config = cfg()

        self.op_mat_obj = opmat()
        self.compute_system = systolic_compute_os()
        self.memory_system = mem_dbsp()

        self.verbose = True

        self.sparsity_ratio_N = 1
        self.sparsity_ratio_M = 1

        # Compression
        self.compression = cp()
        self.original_filter_size = 0
        self.new_filter_size = 0

        # Report items : Compute report
        self.overall_cycles = 0
        self.total_cycles = 0
        self.stall_cycles = 0
        self.num_compute = 0
        self.num_mac_unit = 0
        self.overall_util = 0
        self.mapping_eff = 0
        self.compute_util = 0

        # Report items : BW report
        self.avg_ifmap_sram_bw = 0
        self.avg_filter_sram_bw = 0
        self.avg_ofmap_sram_bw = 0
        self.avg_ifmap_dram_bw = 0
        self.avg_filter_dram_bw = 0
        self.avg_ofmap_dram_bw = 0

        self.avg_filter_metadata_sram_bw = 0

        # Report items : Detailed Access report
        self.ifmap_sram_start_cycle = 0
        self.ifmap_sram_stop_cycle = 0
        self.ifmap_sram_reads = 0

        self.filter_sram_start_cycle = 0
        self.filter_sram_stop_cycle = 0
        self.filter_sram_reads = 0

        self.metadata_reads = 0

        self.ofmap_sram_start_cycle = 0
        self.ofmap_sram_stop_cycle = 0
        self.ofmap_sram_writes = 0

        self.ifmap_dram_start_cycle = 0
        self.ifmap_dram_stop_cycle = 0
        self.ifmap_dram_reads = 0

        self.filter_dram_start_cycle = 0
        self.filter_dram_stop_cycle = 0
        self.filter_dram_reads = 0

        self.ofmap_dram_start_cycle = 0
        self.ofmap_dram_stop_cycle = 0
        self.ofmap_dram_writes = 0

        self.params_set_flag = False
        self.memory_system_ready_flag = False
        self.runs_ready = False
        self.report_items_ready = False

        # Layout Modeling
        self.using_ifmap_custom_layout = True
        self.using_filter_custom_layout = True

    def set_params(self,
                   layer_id=0,
                   config_obj=cfg(), topology_obj=topo(), layout_obj=layout(),
                   verbose=True):
        """
        Method to set the run parameters for housekeeping.
        """

        self.layer_id = layer_id
        self.config = config_obj
        self.topo = topology_obj
        self.layout = layout_obj

        self.op_mat_obj.set_params(layer_id=self.layer_id,
                                   config_obj=self.config,
                                   topoutil_obj=self.topo,
                                   layoututil_obj=self.layout,
                                   )

        self.dataflow = self.config.get_dataflow()
        if self.dataflow == 'os':
            self.compute_system = systolic_compute_os()
        elif self.dataflow == 'ws':
            self.compute_system = systolic_compute_ws()
        elif self.dataflow == 'is':
            self.compute_system = systolic_compute_is()

        arr_dims = self.config.get_array_dims()
        self.using_ifmap_custom_layout = self.config.using_ifmap_custom_layout
        self.using_filter_custom_layout = self.config.using_filter_custom_layout
        if self.using_ifmap_custom_layout:
            print("self.using_ifmap_custom_layout -- True")
        if self.using_filter_custom_layout:
            print("self.using_filter_custom_layout -- True")

        self.num_mac_unit = arr_dims[0] * arr_dims[1]
        self.verbose=verbose

        self.sparsity_ratio_N, self.sparsity_ratio_M = \
            self.topo.get_layer_sparsity_ratio(self.layer_id)

        self.params_set_flag = True

    # This communicates that the memory is being managed externally
    # And the class will not interfere with setting it up
    def set_memory_system(self, mem_sys_obj=mem_dbsp()):
        """
        Method to explicitely set the memory system.
        """
        self.memory_system = mem_sys_obj
        self.memory_system_ready_flag = True

    #
    def calculate_filter_metadata_storage(self, filter_op_mat):
        """
        Method to calculate the storage occupied by the filter matrix and filter metadata when
        sparsity is enabled.
        """
        if self.config.sparsity_support is True:
            sparse_array_np = self.op_mat_obj.sparse_filter_array
            original_storage = 0
            new_storage = 0

            if self.config.sparsity_representation == 'csr':
                original_storage, new_storage, metadata_storage = \
                                                self.compression.get_csr_storage(sparse_array_np)
            elif self.config.sparsity_representation == 'csc':
                original_storage, new_storage, metadata_storage = \
                                                self.compression.get_csc_storage(sparse_array_np)
            elif self.config.sparsity_representation == 'ellpack_block':
                original_storage, new_storage, metadata_storage = \
                                self.compression.get_ellpack_block_storage(sparse_array_np,
                                                                           filter_op_mat,
                                                                           self.sparsity_ratio_M)

            self.original_filter_size += original_storage
            self.new_filter_size += new_storage
            self.metadata_reads += metadata_storage
    # END of metadata calculation

    #
    def run(self):
        """
        Method to run scalesim simulation for a single layer. This method first runs the compute
        part to generate operand, prefetch and demand matrices in order. After that, it runs the
        memory simulation using the demand matrices.
        """
        assert self.params_set_flag, 'Parameters are not set. Run set_params()'

        # 1. Setup and the get the demand from compute system

        # 1.1 Get the operand matrices
        _, ifmap_op_mat = self.op_mat_obj.get_ifmap_matrix()
        _, filter_op_mat = self.op_mat_obj.get_filter_matrix()
        _, ofmap_op_mat = self.op_mat_obj.get_ofmap_matrix()

        # 1.2 Calculate the storage occupied by filter and its metadata
        self.calculate_filter_metadata_storage(filter_op_mat)
        self.num_compute = self.topo.get_layer_num_ofmap_px(self.layer_id) \
                           * self.topo.get_layer_window_size(self.layer_id)

        # 1.3 Get the prefetch matrices for both operands
        if self.dataflow == 'ws':
            self.compute_system.set_params(config_obj=self.config,
                                           ifmap_op_mat=ifmap_op_mat,
                                           filter_op_mat=filter_op_mat,
                                           ofmap_op_mat=ofmap_op_mat,
                                           sparsity_ratio_N=self.sparsity_ratio_N,
                                           sparsity_ratio_M=self.sparsity_ratio_M,
                                           ifmap_op_mat_original=self.op_mat_obj.ifmap_addr_matrix_original,
                                           sparsity_filter_array=self.op_mat_obj.sparse_filter_array)
        else:
            self.compute_system.set_params(config_obj=self.config,
                                           ifmap_op_mat=ifmap_op_mat,
                                           filter_op_mat=filter_op_mat,
                                           ofmap_op_mat=ofmap_op_mat)

        # 1.4 Get the no compute demand matrices from for 2 operands and the output
        ifmap_prefetch_mat, filter_prefetch_mat = self.compute_system.get_prefetch_matrices()

        # 1.4 Get the customed layout for ifmap and filter when it's being specified.
        if self.using_ifmap_custom_layout:
            ifmap_prefetch_mat = self.op_mat_obj.get_ifmap_prefetch_matrix_custom_layout()
        if self.using_filter_custom_layout:
            filter_prefetch_mat = self.op_mat_obj.get_filter_prefetch_matrix_custom_layout()
    
        ifmap_demand_mat, filter_demand_mat, ofmap_demand_mat = self.compute_system.get_demand_matrices()
        #print('DEBUG: Compute operations done')
        # 2. Setup the memory system and run the demands through it to find any memory bottleneck and generate traces


        # 2.1 Setup the memory system if it was not setup externally
        if not self.memory_system_ready_flag:
            word_size = 1           # bytes, this can be incorporated in the config file
            active_buf_frac = 0.5   # This can be incorporated in the config as well

            ifmap_buf_size_kb, filter_buf_size_kb, ofmap_buf_size_kb = self.config.get_mem_sizes()
            ifmap_buf_size_bytes = 1024 * ifmap_buf_size_kb
            filter_buf_size_bytes = 1024 * filter_buf_size_kb
            ofmap_buf_size_bytes = 1024 * ofmap_buf_size_kb

            ifmap_backing_bw = 1
            filter_backing_bw = 1
            ofmap_backing_bw = 1
            estimate_bandwidth_mode = False
            if self.config.use_user_dram_bandwidth():
                bws = self.config.get_bandwidths_as_list()
                ifmap_backing_bw = self.config.ifmap_sram_bank_bandwidth
                filter_backing_bw = self.config.filter_sram_bank_bandwidth
                ofmap_backing_bw = bws[0]

            else:
                arr_row, arr_col = self.config.get_array_dims()
                estimate_bandwidth_mode = True

                # The number 10 elems per cycle is arbitrary
                ifmap_backing_bw = 10
                filter_backing_bw = 10
                ofmap_backing_bw = arr_col

            self.memory_system.set_params(
                    layer_id = self.layer_id,
                    word_size=word_size,
                    ifmap_buf_size_bytes=ifmap_buf_size_bytes,
                    filter_buf_size_bytes=filter_buf_size_bytes,
                    ofmap_buf_size_bytes=ofmap_buf_size_bytes,
                    rd_buf_active_frac=active_buf_frac, wr_buf_active_frac=active_buf_frac,
                    ifmap_backing_buf_bw=ifmap_backing_bw,
                    filter_backing_buf_bw=filter_backing_bw,
                    ofmap_backing_buf_bw=ofmap_backing_bw,
                    verbose=self.verbose,
                    ifmap_sram_bank_num=self.config.ifmap_sram_bank_num,
                    ifmap_sram_bank_port=self.config.ifmap_sram_bank_port,
                    filter_sram_bank_num=self.config.filter_sram_bank_num,
                    filter_sram_bank_port=self.config.filter_sram_bank_port,
                    using_ifmap_custom_layout=self.using_ifmap_custom_layout,
                    using_filter_custom_layout=self.using_filter_custom_layout,
                    estimate_bandwidth_mode=estimate_bandwidth_mode,
                    config=self.config,
                    topo=self.topo
            )

        # 2.2 Install the prefetch matrices to the read buffers to finish setup
        if self.config.use_user_dram_bandwidth() :
            self.memory_system.set_read_buf_prefetch_matrices(
                                                        ifmap_prefetch_mat=ifmap_prefetch_mat,
                                                        filter_prefetch_mat=filter_prefetch_mat
                                                             )
        self.memory_system.service_memory_requests(ifmap_demand_mat,
                                                    filter_demand_mat,
                                                    ofmap_demand_mat)

        self.runs_ready = True

    # This will write the traces
    def save_traces(self, top_path):
        """
        Method to save SRAM and DRAM traces for ifmap, filter and ofmap matrices.
        """
        assert self.params_set_flag, 'Parameters are not set'

        dir_name = top_path + '/layer' + str(self.layer_id)
        if not os.path.isdir(dir_name):
            cmd = 'mkdir ' + dir_name
            os.system(cmd)

        ifmap_sram_filename = dir_name +  '/IFMAP_SRAM_TRACE.csv'
        filter_sram_filename = dir_name + '/FILTER_SRAM_TRACE.csv'
        ofmap_sram_filename = dir_name +  '/OFMAP_SRAM_TRACE.csv'

        ifmap_dram_filename = dir_name +  '/IFMAP_DRAM_TRACE.csv'
        filter_dram_filename = dir_name + '/FILTER_DRAM_TRACE.csv'
        ofmap_dram_filename = dir_name +  '/OFMAP_DRAM_TRACE.csv'

        self.memory_system.print_ifmap_sram_trace(ifmap_sram_filename)
        self.memory_system.print_ifmap_dram_trace(ifmap_dram_filename)
        self.memory_system.print_filter_sram_trace(filter_sram_filename)
        self.memory_system.print_filter_dram_trace(filter_dram_filename)
        self.memory_system.print_ofmap_sram_trace(ofmap_sram_filename)
        self.memory_system.print_ofmap_dram_trace(ofmap_dram_filename)

    #
    def calc_report_data(self):
        """
        Method to generate the report data for a single layer if the run is already completed. This
        method collects and calculates the data for compute, bandwidth and detail reports.
        """
        assert self.runs_ready, 'Runs are not done yet'

        # Compute report
        self.total_cycles = self.memory_system.get_total_compute_cycles()# - self.memory_system.get_stall_cycles()
        self.stall_cycles = self.memory_system.get_stall_cycles()
        self.overall_util = (self.num_compute * 100) / (self.total_cycles * self.num_mac_unit)
        self.mapping_eff = self.compute_system.get_avg_mapping_efficiency() * 100
        self.compute_util = self.compute_system.get_avg_compute_utilization() * 100

        # BW report
        self.ifmap_sram_reads = self.compute_system.get_ifmap_requests()
        self.filter_sram_reads = self.compute_system.get_filter_requests()
        self.ofmap_sram_writes = self.compute_system.get_ofmap_requests()
        self.avg_ifmap_sram_bw = self.ifmap_sram_reads / self.total_cycles

        self.avg_filter_sram_bw = self.filter_sram_reads / self.total_cycles
        self.avg_filter_metadata_sram_bw = self.metadata_reads / self.total_cycles

        self.avg_ofmap_sram_bw = self.ofmap_sram_writes / self.total_cycles

        # Detail report
        self.ifmap_sram_start_cycle, self.ifmap_sram_stop_cycle \
            = self.memory_system.get_ifmap_sram_start_stop_cycles()

        self.filter_sram_start_cycle, self.filter_sram_stop_cycle \
            = self.memory_system.get_filter_sram_start_stop_cycles()

        self.ofmap_sram_start_cycle, self.ofmap_sram_stop_cycle \
            = self.memory_system.get_ofmap_sram_start_stop_cycles()

        self.ifmap_dram_start_cycle, self.ifmap_dram_stop_cycle, self.ifmap_dram_reads \
            = self.memory_system.get_ifmap_dram_details()

        self.filter_dram_start_cycle, self.filter_dram_stop_cycle, self.filter_dram_reads \
            = self.memory_system.get_filter_dram_details()

        self.ofmap_dram_start_cycle, self.ofmap_dram_stop_cycle, self.ofmap_dram_writes \
            = self.memory_system.get_ofmap_dram_details()
        
        self.overall_cycles = int(self.ofmap_dram_stop_cycle - min(self.ifmap_dram_start_cycle,self.filter_dram_start_cycle))
        
        # BW calc for DRAM access
        self.avg_ifmap_dram_bw = self.ifmap_dram_reads / \
                                (self.ifmap_dram_stop_cycle - self.ifmap_dram_start_cycle + 1)
        self.avg_filter_dram_bw = self.filter_dram_reads / \
                                (self.filter_dram_stop_cycle - self.filter_dram_start_cycle + 1)
        self.avg_ofmap_dram_bw = self.ofmap_dram_writes / \
                                (self.ofmap_dram_stop_cycle - self.ofmap_dram_start_cycle + 1)

        self.report_items_ready = True

    #
    def get_layer_id(self):
        """
        Method to return layer id.
        """
        assert self.params_set_flag, 'Parameters are not set yet'
        return self.layer_id

    #
    def get_compute_report_items(self):
        """
        Method to calculate data for the compute report if not already done.
        """
        if not self.report_items_ready:
            self.calc_report_data()

        items = [self.overall_cycles,
                 self.total_cycles,
                 self.stall_cycles,
                 self.overall_util,
                 self.mapping_eff,
                 self.compute_util]
        return items

    #
    def get_bandwidth_report_items(self):
        """
        Method to calculate data for the bandwidth report if not already done.
        """
        if not self.report_items_ready:
            self.calc_report_data()


        if self.config.sparsity_support is True:
            items = [self.avg_ifmap_sram_bw,
                     self.avg_filter_sram_bw,
                     self.avg_filter_metadata_sram_bw,
                     self.avg_ofmap_sram_bw]
        else:
            items = [self.avg_ifmap_sram_bw, self.avg_filter_sram_bw, self.avg_ofmap_sram_bw]
        items += [self.avg_ifmap_dram_bw, self.avg_filter_dram_bw, self.avg_ofmap_dram_bw]

        return items

    #
    def get_detail_report_items(self):
        """
        Method to calculate data for the detailed report if not already done.
        """
        if not self.report_items_ready:
            self.calc_report_data()

        items = [self.ifmap_sram_start_cycle, self.ifmap_sram_stop_cycle, self.ifmap_sram_reads]
        items += [self.filter_sram_start_cycle, self.filter_sram_stop_cycle, self.filter_sram_reads]
        items += [self.ofmap_sram_start_cycle, self.ofmap_sram_stop_cycle, self.ofmap_sram_writes]
        items += [self.ifmap_dram_start_cycle, self.ifmap_dram_stop_cycle, self.ifmap_dram_reads]
        items += [self.filter_dram_start_cycle, self.filter_dram_stop_cycle, self.filter_dram_reads]
        items += [self.ofmap_dram_start_cycle, self.ofmap_dram_stop_cycle, self.ofmap_dram_writes]

        return items

    #
    def get_sparse_report_items(self):
        """
        Method to calculate data for the sparse report if not already done.
        """
        if not self.report_items_ready:
            self.calc_report_data()

        items = [self.original_filter_size, self.new_filter_size, self.metadata_reads]
        items += [self.avg_filter_metadata_sram_bw]

        return items



"""
This file is the main script for running SCALE-Sim with the given topology and configuration files.
It handles argument parsing and execution.
"""

import argparse

from scalesim.scale_sim import scalesim

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-t', metavar='Topology file', type=str,
                        default="./topologies/conv_nets/test.csv",
                        help="Path to the topology file"
                        )
    parser.add_argument('-l', metavar='Layout file', type=str,
                        default="./layouts/conv_nets/test.csv",
                        help="Path to the layout file"
                        )
    parser.add_argument('-c', metavar='Config file', type=str,
                        default="./configs/scale.cfg",
                        help="Path to the config file"
                        )
    parser.add_argument('-p', metavar='log dir', type=str,
                        default="./results/",
                        help="Path to log dir"
                        )
    parser.add_argument('-i', metavar='input type', type=str,
                        default="conv",
                        help="Type of input topology, gemm: MNK, conv: conv"
                        )
    parser.add_argument('-s', metavar='save trace', type=str,
                        default="Y",
                        help="Save Trace: (Y/N)"
                        )

    args = parser.parse_args()
    topology = args.t
    layout = args.l
    config = args.c
    logpath = args.p
    inp_type = args.i
    save_trace = args.s

    GEMM_INPUT = False
    if inp_type == 'gemm':
        GEMM_INPUT = True
    
    if save_trace == 'Y':
        save_space = False
    else:
        save_space = True
   

    s = scalesim(save_disk_space=False,
                 verbose=True,
                 config=config,
                 topology=topology,
                 layout=layout,
                 input_type_gemm=GEMM_INPUT
                 )
    s.run_scale(top_path=logpath)



"""
This module implements the 'systolic_compute_is' class, which simulates a systolic array with Output
Stationary (OS) dataflow. It handles operand prefetching, demand matrix creation, and performance
metrics such as mapping efficiency and compute utilization. It also tracks read and write requests
for IFMAP, Filter, and OFMAP operations.
"""

import math
import numpy as np
from tqdm import tqdm
from scalesim.scale_config import scale_config as cfg


class systolic_compute_os:
    """
    Class that computes the output using Output Stationary dataflow.
    """
    #
    def __init__(self):
        """
        __init__ method.
        """
        # Params set by user
        self.config = cfg()

        self.ifmap_op_mat = np.zeros((1, 1))
        self.ofmap_op_mat = np.zeros((1, 1))
        self.filter_op_mat = np.zeros((1, 1))

        # Derived parameters
        self.Sr = 0
        self.Sc = 0
        self.T = 0

        self.arr_row = 0
        self.arr_col = 0

        self.row_fold = 1
        self.col_fold = 1

        # Generated matrices
        self.ifmap_op_mat_trans = np.zeros((1,1))
        self.ifmap_prefetch_matrix = np.zeros((1,1))
        self.filter_prefetch_matrix = np.zeros((1,1))

        self.ifmap_demand_matrix = np.zeros((1,1))
        self.ofmap_demand_matrix = np.zeros((1,1))
        self.filter_demand_matrix = np.zeros((1,1))

        # Generated metrics
        self.ifmap_reads = 0
        self.filter_reads = 0
        self.ofmap_writes = 0

        self.mapping_efficiency_per_fold = []
        self.compute_utility_per_fold = []

        # Flags
        self.params_set_flag = False
        self.prefetch_mat_ready_flag = False
        self.demand_mat_ready_flag = False

    #
    def set_params(self,
                   config_obj=cfg(),
                   ifmap_op_mat = np.zeros((1,1)),
                   ofmap_op_mat = np.zeros((1,1)),
                   filter_op_mat = np.zeros((1,1))
                ):
        """
        Method to set the output stationary run parameters for housekeeping.
        """

        self.config = config_obj
        self.ifmap_op_mat = ifmap_op_mat
        self.filter_op_mat = filter_op_mat
        self.ofmap_op_mat = ofmap_op_mat

        ifmap_col = self.ifmap_op_mat.shape[1]
        filter_row= self.filter_op_mat.shape[0]

        assert ifmap_col == filter_row, "Dimension mismatch between operands"
        self.ifmap_op_mat_trans = np.transpose(self.ifmap_op_mat)

        self.Sr = self.ifmap_op_mat.shape[0]
        self.Sc = self.filter_op_mat.shape[1]
        self.T = self.ifmap_op_mat.shape[1]

        self.arr_row, self.arr_col = self.config.get_array_dims()

        self.row_fold = math.ceil(self.Sr / self.arr_row)
        self.col_fold = math.ceil(self.Sc / self.arr_col)

        self.params_set_flag = True

    #
    def create_prefetch_matrices(self):
        """
        Method to create ifmap and filter prefetch matrices. These matrices are prefetched in the
        SRAM before running memory simulation.
        """
        assert self.params_set_flag, 'Parameters are not set'

        self.create_ifmap_prefetch_mat()
        self.create_filter_prefetch_mat()

        self.prefetch_mat_ready_flag = True

    #
    def create_ifmap_prefetch_mat(self):
        """
        Method to create ifmap prefetch matrix.
        """
        assert self.params_set_flag, 'Parameters are not set'

        for fr in range(self.row_fold):
            start_row_idx = fr * self.arr_row
            end_row_idx = min(start_row_idx + self.arr_row, self.Sr)

            delta = self.arr_row - (end_row_idx - start_row_idx)

            # The usage of row idx in cols is correct as this is the transposed matrix
            # Thus, Sr is along the cols and T is along the rows of this matrix
            # This is how the traces will be generated as well
            this_fold_prefetch = self.ifmap_op_mat_trans[:,start_row_idx: end_row_idx]

            #If there is under utilization, fill them with null requests
            if delta > 0:
                null_req_mat = np.ones((self.T, delta)) * -1
                this_fold_prefetch = np.concatenate((this_fold_prefetch, null_req_mat), axis=1)

            if fr == 0:
                self.ifmap_prefetch_matrix = this_fold_prefetch
            else:
                self.ifmap_prefetch_matrix = \
                    np.concatenate((self.ifmap_prefetch_matrix, this_fold_prefetch), axis=0)

        # Fixing ISSUE #15, #16
        # Roll out the matrices along the diagonal to account for temporal locality when there is a
        # skew in demand
        #print('DEBUG: create_ifmap_prefetch_mat()')
        #start_time = time.time()

        M, N = self.ifmap_prefetch_matrix.shape
        num_elems = M * N
        num_diags = M + N
        prefetches = np.zeros((1,num_elems))
        idx = 0

        pbar = tqdm(total=M*N, disable=True)
        #print('DEBUG: Total = ' + str(num_elems) + ' Diags = ' + str(num_diags))

        for diag_id in range(num_diags):
            max_row_id = min(diag_id, M - 1)
            min_row_id = max(0, diag_id - N + 1)
            valid_rows = max_row_id - min_row_id + 1

            for offset in range(valid_rows):
                row_id = max_row_id - offset
                col_id = diag_id - row_id

                elem = self.ifmap_prefetch_matrix[row_id][col_id]
                prefetches[0, idx] = elem
                idx += 1
                pbar.update(1)

        pbar.close()
        self.ifmap_prefetch_matrix = prefetches

        #t = time.time() - start_time
        #print('DEBUG: create_ifmap_prefetch_mat =' + str(t))

    #
    def create_filter_prefetch_mat(self):
        """
        Method to create filter prefetch matrix.
        """
        assert self.params_set_flag, 'Parameters are not set'

        for fc in range(self.col_fold):
            col_start_id = fc * self.arr_col
            col_end_id = min(col_start_id + self.arr_col, self.Sc)

            delta = self.arr_col - (col_end_id - col_start_id)

            this_fold_prefetch = self.filter_op_mat[:,col_start_id:col_end_id]

            if delta > 0:
                null_req_mat = np.ones((self.T, delta)) * -1
                this_fold_prefetch = np.concatenate((this_fold_prefetch, null_req_mat), axis=1)

            if fc == 0:
                self.filter_prefetch_matrix = this_fold_prefetch
            else:
                self.filter_prefetch_matrix = \
                    np.concatenate((self.filter_prefetch_matrix, this_fold_prefetch), axis=0)

        # Fixing ISSUE #15, #16
        # Roll out the matrices along the diagonal to account for temporal locality when there is a
        # skew in demand
        #print('DEBUG: create_filter_prefetch_mat()')
        #start_time = time.time()

        M, N = self.filter_prefetch_matrix.shape
        num_elems = M * N
        num_diags = M + N
        prefetches = np.zeros((1, num_elems))
        idx = 0

        pbar = tqdm(total=M * N, disable=True)
        # print('DEBUG: Total = ' + str(num_elems) + ' Diags = ' + str(num_diags))

        for diag_id in range(num_diags):
            max_row_id = min(diag_id, M - 1)
            min_row_id = max(0, diag_id - N + 1)
            valid_rows = max_row_id - min_row_id + 1

            for offset in range(valid_rows):
                row_id = max_row_id - offset
                col_id = diag_id - row_id

                elem = self.filter_prefetch_matrix[row_id][col_id]
                prefetches[0, idx] = elem
                idx += 1
                pbar.update(1)

        pbar.close()
        self.filter_prefetch_matrix = prefetches

        #t = time.time() - start_time
        #print('DEBUG: create_filter_prefetch_mat =' + str(t))

    #
    def create_demand_matrices(self):
        """
        Method to create ifmap, filter and ofmap demand matrices from the operand matrices. They
        contain several folds of ifmap, filter and ofmap demands. The folding happens because
        operand matrices are generally larger than systolic array dimensions.
        """
        assert self.params_set_flag, 'Parameters are not set'

        self.create_ifmap_demand_mat()
        self.create_filter_demand_mat()
        self.create_ofmap_demand_mat()

        assert self.ifmap_demand_matrix.shape[0] == self.filter_demand_matrix.shape[0], \
               'IFMAP and Filter demands out of sync'
        assert self.ofmap_demand_matrix.shape[0] == self.filter_demand_matrix.shape[0], \
               'OFMAP and Filter demands out of sync'
        assert self.ifmap_demand_matrix.shape[1] == self.arr_row, 'IFMAP demands exceed the rows'
        assert self.filter_demand_matrix.shape[1] == self.arr_col,'Filter demands exceed the cols'
        assert self.ofmap_demand_matrix.shape[1] == self.arr_col, 'OFMAP demands exceed the cols'

        self.demand_mat_ready_flag = True

    #
    def create_ifmap_demand_mat(self):
        """
        Method to create ifmap demand matrix.
        """
        assert self.params_set_flag, 'Parameters are not set'

        # Anand: Concatenation issue fix
        inter_fold_gap_suffix = self.arr_col - 1
        inter_fold_gap_suffix_mat = np.ones((inter_fold_gap_suffix, self.arr_row)) * -1

        # DEBUG section
        #print('DEBUG: create_ifmap_demand_mat()')
        pbar = tqdm(total=self.col_fold * self.row_fold, disable=True)

        for fc in range(self.col_fold):
            for fr in range(self.row_fold):
                row_start_id = fr * self.arr_row
                row_end_idx = min(row_start_id + self.arr_row, self.Sr)
                delta = self.arr_row - (row_end_idx - row_start_id)

                # Indexing the cols with row start and row end idx are correct
                # See the comment on ifmap_prefetch generation
                this_fold_demand = self.ifmap_op_mat_trans[:,row_start_id: row_end_idx]
                self.ifmap_reads += this_fold_demand.shape[0] * this_fold_demand.shape[1]

                # Take into account under utilization
                if delta > 0:
                    null_req_mat = np.ones((self.T, delta)) * -1
                    this_fold_demand = np.concatenate((this_fold_demand, null_req_mat), axis=1)

                # In this computation scheme we are allowing the generated outputs to drain out
                # before starting the next fold
                # This portion accounts for that extra time by adding null requests
                this_fold_demand = np.concatenate((this_fold_demand, inter_fold_gap_suffix_mat),
                                                  axis=0)

                # Add skew to the IFMAP demand matrix to reflect systolic pipeline fill
                this_fold_demand = skew_matrix(this_fold_demand)

                if fr == 0 and fc == 0:
                    self.ifmap_demand_matrix = this_fold_demand
                else:
                    self.ifmap_demand_matrix = \
                        np.concatenate((self.ifmap_demand_matrix, this_fold_demand), axis=0)

                pbar.update(1)

        pbar.close()
        # TODO: cleanup
        # Add skew to the IFMAP demand matrix to reflect systolic pipeline fill
        #self.ifmap_demand_matrix = skew_matrix(self.ifmap_demand_matrix)

    #
    def create_filter_demand_mat(self):
        """
        Method to create filter demand matrix.
        """
        assert self.params_set_flag, 'Parameters are not set'

        inter_fold_gap_suffix = self.arr_row - 1
        inter_fold_gap_suffix_mat = np.ones((inter_fold_gap_suffix, self.arr_col)) * -1

        # Debug messages
        #print('DEBUG: create_filter_demand_mat()')
        pbar = tqdm(total=self.col_fold * self.row_fold, disable=True)

        for fc in range(self.col_fold):
            for fr in range(self.row_fold):
                col_start_id = fc * self.arr_col
                col_end_idx = min(col_start_id + self.arr_col, self.Sc)
                delta = self.arr_col - (col_end_idx - col_start_id)

                this_fold_demand = self.filter_op_mat[:, col_start_id: col_end_idx]
                self.filter_reads += this_fold_demand.shape[0] * this_fold_demand.shape[1]

                # Take into account under utilization
                if delta > 0:
                    null_req_mat = np.ones((self.T, delta)) * -1
                    this_fold_demand = np.concatenate((this_fold_demand, null_req_mat), axis=1)

                # In this computation scheme we are allowing the generated outputs to drain out
                # before starting the next fold
                # This portion accounts for that extra time by adding null requests
                this_fold_demand = np.concatenate((this_fold_demand, inter_fold_gap_suffix_mat),
                                                  axis=0)

                # Add skew to the Filter demand matrix to reflect systolic pipeline fill
                this_fold_demand = skew_matrix(this_fold_demand)

                if fr == 0 and fc == 0:
                    self.filter_demand_matrix = this_fold_demand
                else:
                    self.filter_demand_matrix = \
                        np.concatenate((self.filter_demand_matrix, this_fold_demand), axis=0)

                pbar.update(1)

        pbar.close()
        # TODO: Cleanup
        # Add skew to the Filter demand matrix to reflect systolic pipeline fill
        #self.filter_demand_matrix = skew_matrix(self.filter_demand_matrix)

    #
    def create_ofmap_demand_mat(self):
        """
        Method to create ofmap demand matrix.
        """
        assert self.params_set_flag, 'Parameters are not set'

        inter_fold_gap_prefix = self.T  - 1
        inter_fold_gap_prefix_mat = np.ones((inter_fold_gap_prefix, self.arr_col)) * -1

        # Debug messages
        #print('DEBUG: create_ifmap_demand_mat()')
        pbar = tqdm(total=self.col_fold * self.row_fold, disable=True)

        for fc in range(self.col_fold):
            for fr in range(self.row_fold):
                row_start_id = fr * self.arr_row
                row_end_idx = min(row_start_id + self.arr_row, self.Sr)
                row_delta = self.arr_row - (row_end_idx - row_start_id)

                col_start_id = fc * self.arr_col
                col_end_idx = min(col_start_id + self.arr_col, self.Sc)
                col_delta = self.arr_col - (col_end_idx - col_start_id)

                this_fold_demand = \
                    self.ofmap_op_mat[row_start_id: row_end_idx, col_start_id: col_end_idx]
                self.ofmap_writes += this_fold_demand.shape[0] * this_fold_demand.shape[1]

                # Adding null requests when there is under utilization ie. no mapping along a few
                # rows or cols
                if col_delta > 0:
                    null_req_mat = np.ones((this_fold_demand.shape[0], col_delta)) * -1
                    this_fold_demand = np.concatenate((this_fold_demand, null_req_mat), axis=1)

                if row_delta > 0:
                    null_req_mat = np.ones((row_delta, self.arr_col)) * -1
                    this_fold_demand = np.concatenate((this_fold_demand, null_req_mat), axis=0)

                # Reflect along the rows
                # This is a characteristic of the fact that the outputs are streamed out from the
                # bottom edge.
                # If the outputs are streamed out from the top edge instead, then this step is not
                # needed.
                this_fold_demand = np.flip(this_fold_demand, 0)
                self.ofmap_writes += this_fold_demand.shape[0] + this_fold_demand.shape[1]

                # Now add the prefix matrix
                # These are the null demands to account for when the operands are streamed in
                # and the OFMAPS are not ready
                this_fold_demand = np.concatenate((inter_fold_gap_prefix_mat, this_fold_demand),
                                                  axis=0)

                # Calculate the mapping efficiency
                row_used = min(self.arr_row, row_end_idx - row_start_id)
                col_used = min(self.arr_col, col_end_idx - col_start_id)
                mac_used = row_used * col_used
                mapping_eff_this_fold = mac_used / (self.arr_row * self.arr_col)

                cycles_this_fold = this_fold_demand.shape[0] + this_fold_demand.shape[1] - 1
                compute_cycles_this_fold = mac_used * self.T
                compute_util_this_fold = \
                    compute_cycles_this_fold / (self.arr_row * self.arr_col * cycles_this_fold)

                self.mapping_efficiency_per_fold.append(mapping_eff_this_fold)
                self.compute_utility_per_fold.append(compute_util_this_fold)

                # Add skew to the OFMAP demand matrix to reflect systolic pipeline fill
                this_fold_demand = skew_matrix(this_fold_demand)

                if fr == 0 and fc == 0:
                    self.ofmap_demand_matrix = this_fold_demand
                else:
                    self.ofmap_demand_matrix = \
                        np.concatenate((self.ofmap_demand_matrix, this_fold_demand), axis=0)

                pbar.update(1)

        pbar.close()
        # TODO: cleanup
        # Add skew to the OFMAP demand matrix to reflect systolic pipeline fill
        #self.ofmap_demand_matrix = skew_matrix(self.ofmap_demand_matrix)

    #
    def get_ifmap_prefetch_mat(self):
        """
        Method to get ifmap prefetch matrix.
        """
        if not self.prefetch_mat_ready_flag:
            self.create_prefetch_matrices()

        return self.ifmap_prefetch_matrix

    #
    def get_filter_prefetch_mat(self):
        """
        Method to get filter prefetch matrix.
        """
        if not self.prefetch_mat_ready_flag:
            self.create_prefetch_matrices()

        return self.filter_prefetch_matrix

    #
    def get_prefetch_matrices(self):
        """
        Method to get ifmap and filter prefetch matrices.
        """
        if not self.prefetch_mat_ready_flag:
            self.create_prefetch_matrices()

        return self.ifmap_prefetch_matrix, self.filter_prefetch_matrix

    #
    def get_ifmap_demand_mat(self):
        """
        Method to get ifmap demand matrix.
        """
        if not self.demand_mat_ready_flag:
            self.create_demand_matrices()

        return self.ifmap_demand_matrix

    #
    def get_filter_demand_mat(self):
        """
        Method to get filter demand matrix.
        """
        if not self.demand_mat_ready_flag:
            self.create_demand_matrices()

        return self.filter_demand_matrix

    #
    def get_ofmap_demand_mat(self):
        """
        Method to get ofmap demand matrix.
        """
        if not self.demand_mat_ready_flag:
            self.create_demand_matrices()

        return self.ofmap_demand_matrix

    #
    def get_demand_matrices(self):
        """
        Method to get ifmap, filter and ofmap demand matrices.
        """
        if not self.demand_mat_ready_flag:
            self.create_demand_matrices()

        return self.ifmap_demand_matrix, self.filter_demand_matrix, self.ofmap_demand_matrix

    #
    def get_avg_mapping_efficiency(self):
        """
        Method to get average mapping efficincy on the systolic array.
        """
        assert self.demand_mat_ready_flag, 'Computes not ready yet'

        agg = sum(self.mapping_efficiency_per_fold)
        num = len(self.mapping_efficiency_per_fold)

        avg_mapping_eff = agg / num

        return avg_mapping_eff

    #
    def get_avg_compute_utilization(self):
        """
        Method to get average compute utilization on the systolic array.
        """
        assert self.demand_mat_ready_flag, 'Computes not ready yet'

        agg = sum(self.compute_utility_per_fold)
        num = len(self.compute_utility_per_fold)

        avg_compute_util = agg / num

        return avg_compute_util

    #
    def get_ifmap_requests(self):
        """
        Method to get ifmap read requests.
        """
        assert self.demand_mat_ready_flag, 'Computes not ready yet'
        return self.ifmap_reads

    #
    def get_filter_requests(self):
        """
        Method to get filter read requests.
        """
        assert self.demand_mat_ready_flag, 'Computes not ready yet'
        return self.filter_reads

    #
    def get_ofmap_requests(self):
        """
        Method to get ofmap write requests.
        """
        assert self.demand_mat_ready_flag, 'Computes not ready yet'
        return self.ofmap_writes

#
def skew_matrix(input_matrix_np):
    """
    Method to add skew to the input matix to maintain systolic array flow.
    Example:
        Input matrix:
        1 1 1 1 1 1 1 1 1

        Output matrix:
            1 1 1
          1 1 1
        1 1 1
    """
    rows, cols = input_matrix_np.shape

    out_matrix_np = np.full((rows + cols - 1, cols), -1, dtype=input_matrix_np.dtype)

    for c in range(cols):
        out_matrix_np[c:c + rows, c] = input_matrix_np[:, c]

    return out_matrix_np



"""
External DRAM write requests serviced by Ramulator
"""
import numpy as np
from scalesim.scale_config import scale_config as config
from bisect import bisect_left

# This is shell module to ensure continuity

class write_port:
    """
    Class to define external DRAM memory requests with Ramulator
    """
    #
    def __init__(self):
        """
        __init__ module.
        """
        self.latency = 0
        self.ramulator_trace = False
        self.latency_matrix = []
        self.bw = 10
        self.request_queue_size = 100
        self.request_queue_status = 0
        self.stall_cycles = 0
        self.request_array = []
        self.count = 0
        self.config = config()
    
    def def_params( self,
                    config = config(),
                    latency_file =''
                ):
        """
        Method to define the paths of ramulator trace numpy files 
        and write request queue sizes.
        """
        self.config = config
        self.ramulator_trace = self.config.get_ramulator_trace()
        self.request_queue_size = self.config.get_req_buf_sz_wr()
        self.bw = self.config.get_bandwidths_as_list()[0]
        if self.ramulator_trace == True:
            self.latency_matrix = np.load(latency_file)
        self.latency=0
    #

    def find_latency(self):
        """
        Method to map DRAM return path latency for each transactions.
        """
        if(self.count < len(self.latency_matrix)):
            latency_out = self.latency_matrix[self.count]
            #print(str(self.count)+ ' ' + str(latency_out))
            self.count+=1
        else:
            latency_out = self.latency
        if(latency_out > 10000):
            latency_out = 0

        return latency_out

    def service_writes(self, incoming_requests_arr_np, incoming_cycles_arr_np):
        """
        Method to service read request by the read buffer.
        Check for hit in the request queue or add the DRAM
        roundtrip latency for each transaction reported by 
        Ramulator.
        """
        if self.ramulator_trace == False:
            out_cycles_arr_np = incoming_cycles_arr_np + self.latency
            out_cycles_arr_np = out_cycles_arr_np.reshape((out_cycles_arr_np.shape[0], 1))
            return out_cycles_arr_np

        updated_req_timestamp = incoming_cycles_arr_np[0][0]
        print(updated_req_timestamp)
        out_cycles_arr = np.zeros(incoming_cycles_arr_np.shape[0])
        for i in range(len(incoming_cycles_arr_np)):
            out_cycles_arr[i] = incoming_cycles_arr_np[i][0] + self.stall_cycles + self.find_latency()
            self.request_array.append(out_cycles_arr[i])
            if len(self.request_array) == self.request_queue_size:
                updated_req_timestamp = incoming_cycles_arr_np[i][0] + self.stall_cycles
                self.request_array.sort()
                if self.request_array[0] >= updated_req_timestamp:
                    self.stall_cycles += self.request_array[0] - updated_req_timestamp
                    updated_req_timestamp = self.request_array[0]
                    self.request_array.pop(0)
                else:
                    index = bisect_left(self.request_array,updated_req_timestamp)
                    if index == len(self.request_array):
                        self.request_array = []
                        print("Empty array")
                    else:
                        self.request_array = self.request_array[index:]
            elif len(self.request_array) > self.request_queue_size:
                self.request_array = self.request_array[-self.request_queue_size:]
        # --- Placeholder for zeroization logic
        #if flush == 0:
        #    self.stall_cycles = 0

        self.stall_cycles =0
        return out_cycles_arr



"""
This file contains the class 'ScalesimReport' containing the utility functions to read the report
data.
"""

import pandas as pd


class ScalesimReport:
    """
    Class which has the utility functions to load report data.
    """
    def __init__(self):
        """
        __init__ method.
        """
        self.compute_df = pd.DataFrame()
        self.bandwidths_df = pd.DataFrame()
        self.details_df = pd.DataFrame()

        self.compute_df_ready = False
        self.bandwidths_df_ready = False
        self.details_df_ready = False

    #
    def load_data(self, data_dir='.', run_name=''):
        """
        Method to load data of compute, bandwidth and detail reports.
        """
        self.load_compute_report_data(data_dir=data_dir, run_name=run_name)
        self.load_bandwidth_report_data(data_dir=data_dir, run_name=run_name)
        self.load_detail_report_data(data_dir=data_dir, run_name=run_name)

    #
    def load_compute_report_data(self, data_dir='.', run_name=''):
        """
        Method to load data of the compute report.
        """
        csv_filename = data_dir + '/' + run_name + '/COMPUTE_REPORT.csv'
        self.compute_df = pd.read_csv(csv_filename, sep=r'\s*,\s*', engine='python')
        self.compute_df_ready = True

    #
    def load_bandwidth_report_data(self, data_dir='.', run_name=''):
        """
        Method to load data of the bandwidth report.
        """
        csv_filename = data_dir + '/' + run_name + '/BANDWIDTH_REPORT.csv'
        self.bandwidths_df = pd.read_csv(csv_filename, sep=r'\s*,\s*', engine='python')
        self.bandwidths_df_ready = True

    #
    def load_detail_report_data(self, data_dir='.', run_name=''):
        """
        Method to load data of the detail report.
        """
        csv_filename = data_dir + '/' + run_name + '/DETAILED_ACCESS_REPORT.csv'
        self.bandwidths_df = pd.read_csv(csv_filename, sep=r'\s*,\s*', engine='python')
        self.details_df_ready = True

    #
    def get_total_cycles_single_layer(self, layer_id=0):
        """
        Method to get total cycles of a single layer of the workload.
        """
        assert self.compute_df_ready, 'Data not read yet'
        col_name = 'Total Cycles'
        elem = self.compute_df[col_name][layer_id]

        return elem

    #
    def get_compute_cycles_all_layer(self):
        """
        Method to get total cycles of all the layers of the workload.
        """
        assert self.compute_df_ready, 'Data not read yet'
        col_name = 'Total Cycles'
        elems = self.compute_df[col_name].to_list()

        return elems

    #
    def get_sram_ifmap_reads_single_layer(self, layer_id=0):
        """
        Method to get ifmap reads of a single layer of the workload.
        """
        assert self.details_df_ready, 'Data not read yet'
        col_name = 'SRAM IFMAP Reads'
        elem = self.details_df[col_name][layer_id]

        return elem

    #
    def get_sram_ifmap_reads_all_layer(self):
        """
        Method to get ifmap reads of all the layers of the workload.
        """
        assert self.details_df_ready, 'Data not read yet'
        col_name = 'SRAM IFMAP Reads'
        elems = self.details_df[col_name].to_list()

        return elems

if __name__ == '__main__':
    rpt = ScalesimReport()
    rpt.load_data(data_dir='./tutorial1_runs', run_name='scale_sim_tutorial1_64x64_os')

    print(rpt.get_compute_cycles_all_layer())
    print(rpt.get_total_cycles_single_layer(layer_id=1))



"""
The `write_buffer` class simulates the memory operations of the OFMAP SRAM for a double-buffered
write system in systolic array-based computations.
"""
# TODO: Verification Pending
import time
import math
import numpy as np
# import matplotlib.pyplot as plt
from tqdm import tqdm
from scalesim.memory.write_port import write_port


class write_buffer:
    """
    Class which runs the memory simulation of the OFMAP SRAM.
    """
    #
    def __init__(self):
        """
        __init__ method.
        """
        # Buffer properties: User specified
        self.total_size_bytes = 128
        self.word_size = 1
        self.active_buf_frac = 0.9

        # Buffer properties: Calculated
        self.total_size_elems = math.floor(self.total_size_bytes / self.word_size)
        self.active_buf_size = int(math.ceil(self.total_size_elems * self.active_buf_frac))
        self.drain_buf_size = self.total_size_elems - self.active_buf_size

        # Backing interface properties
        self.backing_buffer = write_port()
        self.req_gen_bandwidth = 100

        # Status of the buffer
        self.free_space = self.total_size_elems
        self.drain_buf_start_line_id = 0
        self.drain_buf_end_line_id = 0

        # Helper data structures for faster execution
        self.line_idx = 0
        self.current_line = np.ones((1, 1)) * -1
        self.max_cache_lines = 2 ** 10  # TODO: This is arbitrary, check if this can be tuned
        self.trace_matrix_cache = np.zeros((1, 1))

        # Access counts
        self.num_access = 0

        # Trace matrix
        self.trace_matrix = np.zeros((1, 1))
        self.cycles_vec = np.zeros((1, 1))

        # Flags
        # This variable determines where the new requests should be buffered
        # 0: Directly in the drain buffer
        # 1: In the active buffer, while the drain buffer is flushed
        self.state = 0
        self.drain_end_cycle = 0

        self.trace_valid = False
        # Fixing ISSUE #10
        self.trace_matrix_cache_empty = True
        self.trace_matrix_empty = True

    #
    def set_params(self, backing_buf_obj,
                   total_size_bytes=128, word_size=1, active_buf_frac=0.9,
                   backing_buf_bw=100
                   ):
        """
        Method to set the ofmap memory simulation parameters for housekeeping.
        """
        self.total_size_bytes = total_size_bytes
        self.word_size = word_size

        assert 0.5 <= active_buf_frac < 1, "Valid active buf frac [0.5,1)"
        self.active_buf_frac = active_buf_frac

        self.backing_buffer = backing_buf_obj
        self.req_gen_bandwidth = backing_buf_bw

        self.total_size_elems = math.floor(self.total_size_bytes / self.word_size)
        self.active_buf_size = int(math.ceil(self.total_size_elems * self.active_buf_frac))
        self.drain_buf_size = self.total_size_elems - self.active_buf_size
        self.free_space = self.total_size_elems

    #
    def reset(self):
        """
        Method to reset the write buffer parameters.
        """
        self.total_size_bytes = 128
        self.word_size = 1
        self.active_buf_frac = 0.9

        self.backing_buffer = write_buffer()
        self.req_gen_bandwidth = 100

        self.free_space = self.total_size_elems
        self.drain_end_cycle = 0

        self.trace_matrix = np.zeros((1, 1))

        self.num_access = 0
        self.state = 0

        self.trace_valid = False
        # Fixing ISSUE #10
        self.trace_matrix_cache_empty = True
        self.trace_matrix_empty = True

    #
    def store_to_trace_mat_cache(self, elem):
        """
        Method to add the incoming element to the trace matrix cache.
        """
        if elem == -1:
            return

        if self.current_line.shape == (1,1):    # This line is empty
            self.current_line = np.ones((1, self.req_gen_bandwidth)) * -1

        self.current_line[0, self.line_idx] = elem
        self.line_idx += 1
        self.free_space -= 1

        if not self.line_idx < self.req_gen_bandwidth:
            # Store to the cache matrix
            # Fixing ISSUE #10
            # if self.trace_matrix_cache.shape == (1,1):
            if self.trace_matrix_cache_empty:
                self.trace_matrix_cache = self.current_line
                self.trace_matrix_cache_empty = False
            else:
                self.trace_matrix_cache = np.concatenate((
                                                          self.trace_matrix_cache,
                                                          self.current_line
                                                         ),
                                                         axis=0)

            self.current_line = np.ones((1,1)) * -1
            self.line_idx = 0

            if not self.trace_matrix_cache.shape[0] < self.max_cache_lines:
                self.append_to_trace_mat()

    #
    def append_to_trace_mat(self, force=False):
        """
        Method to append to the trace matrix cache.
        """
        if force:
            # This forces the contents for self.current_line and self.trace_matrix
            # cache to be dumped
            if not self.line_idx == 0:
                #if self.trace_matrix_cache.shape == (1,1):
                if self.trace_matrix_cache_empty:
                    self.trace_matrix_cache = self.current_line
                    self.trace_matrix_cache_empty = False
                else:
                    self.trace_matrix_cache = np.concatenate((
                                                              self.trace_matrix_cache,
                                                              self.current_line
                                                             ),
                                                             axis=0)

                self.current_line = np.ones((1,1)) * -1
                self.line_idx = 0
        # Fixing ISSUE #10
        # if self.trace_matrix_cache.shape == (1,1):
        if self.trace_matrix_cache_empty:
            return

        #if self.trace_matrix.shape == (1,1):
        if self.trace_matrix_empty:
            self.trace_matrix = self.trace_matrix_cache
            self.drain_buf_start_line_id = 0
            self.trace_matrix_empty = False
        else:
            self.trace_matrix = np.concatenate((self.trace_matrix, self.trace_matrix_cache), axis=0)

        self.trace_matrix_cache = np.zeros((1,1))
        # Fixing ISSUE #10
        self.trace_matrix_cache_empty = True

    #
    def service_writes(self, incoming_requests_arr_np, incoming_cycles_arr_np):
        """
        Method to service write requests coming from systolic array. Logic: Assuming no miss, keep
        adding to the active buffer. Once the active buffer is full, drain a part of it to the DRAM.
        """
        assert incoming_cycles_arr_np.shape[0] == incoming_requests_arr_np.shape[0], \
                                                  'Cycles and requests do not match'
        out_cycles_arr = []
        offset = 0

        # DEBUG_num_drains = 0
        # DEBUG_append_to_trace_times = []

        for i in tqdm(range(incoming_requests_arr_np.shape[0]), disable=True):
            row = incoming_requests_arr_np[i]
            cycle = incoming_cycles_arr_np[i]
            current_cycle = cycle[0] + offset

            for elem in row:
                # Pay no attention to empty requests
                if elem == -1:
                    continue

                self.store_to_trace_mat_cache(elem)

                if current_cycle < self.drain_end_cycle:
                    if not self.free_space > 0:
                        offset += max(self.drain_end_cycle - current_cycle, 0)
                        current_cycle = self.drain_end_cycle

                elif self.free_space < (self.total_size_elems - self.drain_buf_size):
                    self.append_to_trace_mat(force=True)
                    self.drain_end_cycle = self.empty_drain_buf(empty_start_cycle=current_cycle)
                    # TODO sarbartha
                    #current_cycle = self.drain_end_cycle

            out_cycles_arr.append(current_cycle)

        num_lines = incoming_requests_arr_np.shape[0]
        out_cycles_arr_np = np.asarray(out_cycles_arr).reshape((num_lines, 1))

        #print('DEBUG: Num Drains = ' + str(DEBUG_num_drains))
        #print('DEBUG: Num appeneds = ' + str(len(DEBUG_append_to_trace_times)))
        #plt.plot(DEBUG_append_to_trace_times)
        #plt.show()

        return out_cycles_arr_np

    #
    def empty_drain_buf(self, empty_start_cycle=0):
        """
        Method to drain the drain buffer once the active buffer is full.
        """

        lines_to_fill_dbuf = int(math.ceil(self.drain_buf_size / self.req_gen_bandwidth))
        self.drain_buf_end_line_id = self.drain_buf_start_line_id + lines_to_fill_dbuf
        self.drain_buf_end_line_id = min(self.drain_buf_end_line_id, self.trace_matrix.shape[0])

        requests_arr_np = \
                    self.trace_matrix[self.drain_buf_start_line_id: self.drain_buf_end_line_id, :]
        num_lines = requests_arr_np.shape[0]

        data_sz_to_drain = num_lines * requests_arr_np.shape[1]
        # Adjust for -1
        for elem in requests_arr_np[-1,:]:
            if elem == -1:
                data_sz_to_drain -= 1
        self.num_access += data_sz_to_drain

        cycles_arr = [x+empty_start_cycle for x in range(num_lines)]
        cycles_arr_np = np.asarray(cycles_arr).reshape((num_lines, 1))
        serviced_cycles_arr = self.backing_buffer.service_writes(requests_arr_np, cycles_arr_np)

        # Assign the cycles vector which will be used to generate the complete trace
        if not self.trace_valid:
            self.cycles_vec = serviced_cycles_arr
            self.trace_valid = True
        else:
            self.cycles_vec = np.concatenate((self.cycles_vec, serviced_cycles_arr), axis=0)

        service_end_cycle = np.amax(serviced_cycles_arr)
        self.free_space += data_sz_to_drain

        self.drain_buf_start_line_id = self.drain_buf_end_line_id
        return service_end_cycle

    #
    def empty_all_buffers(self, cycle):
        """
        Method to drain all of the active buffer.
        """
        self.append_to_trace_mat(force=True)

        if self.trace_matrix_empty:
            return

        while self.drain_buf_start_line_id < self.trace_matrix.shape[0]:
            self.drain_end_cycle = self.empty_drain_buf(empty_start_cycle=cycle)
            cycle = self.drain_end_cycle + 1

    #
    def get_trace_matrix(self):
        """
        Method to get the write buffer trace matrix. It contains addresses requsted by the systolic
        array and the cycles (first column) at which the requests are made.
        """
        if not self.trace_valid:
            print('No trace has been generated yet')
            return

        trace_matrix = np.column_stack((self.cycles_vec, self.trace_matrix))

        return trace_matrix

    #
    def get_free_space(self):
        """
        Method to get free space of the write buffer.
        """
        return self.free_space

    #
    def get_num_accesses(self):
        """
        Method to get number of accesses of the write buffer if trace_valid flag is set.
        """
        assert self.trace_valid, 'Traces not ready yet'
        return self.num_access

    #
    def get_external_access_start_stop_cycles(self):
        """
        Method to get start and stop cycles of the write buffer if trace_valid flag is set.
        """
        assert self.trace_valid, 'Traces not ready yet'
        start_cycle = np.amin(self.cycles_vec)
        end_cycle = np.amax(self.cycles_vec)
        return start_cycle, end_cycle

    #
    def print_trace(self, filename):
        """
        Method to write the write buffer trace matrix to a file.
        """
        if not self.trace_valid:
            print('No trace has been generated yet')
            return
        trace_matrix = self.get_trace_matrix()
        np.savetxt(filename, trace_matrix, fmt='%s', delimiter=",")



"""
Double buffer read memory implementation
"""
# TODO: Verification Pending
import math
import numpy as np
from tqdm import tqdm

from scalesim.memory.read_port import read_port


class read_buffer:
    """
    Class which runs the memory simulation of double buffered ifmap/filter SRAM. The double
    buffering helps to hide the DRAM latency when the SRAM is servicing requests from the systolic
    array using one of the buffers while the other buffer prefetches from the DRAM.
    """
    #
    def __init__(self):
        """
        __init__ method.
        """
        # Buffer properties: User specified
        self.total_size_bytes = 128
        self.word_size = 1    # Bytes
        self.active_buf_frac = 0.9
        self.hit_latency = 1  # Cycles after which a request is served if already in the buffer

        # Buffer properties: Calculated
        self.total_size_elems = math.floor(self.total_size_bytes / self.word_size)
        self.active_buf_size = int(math.ceil(self.total_size_elems * 0.9))
        self.prefetch_buf_size = self.total_size_elems - self.active_buf_size

        # Backing interface properties
        self.backing_buffer = read_port()
        self.req_gen_bandwidth = 100            # words per cycle

        # Status of the buffer
        self.hashed_buffer = {}
        self.num_lines = 0
        self.num_active_buf_lines = 1
        self.num_prefetch_buf_lines = 1
        self.active_buffer_set_limits = []
        self.prefetch_buffer_set_limits = []

        # Variables to enable prefetching
        self.fetch_matrix = np.ones((1, 1))
        self.last_prefetch_cycle = -1
        self.next_line_prefetch_idx = 0
        self.next_col_prefetch_idx = 0

        # Access counts
        self.num_access = 0

        # Trace matrix
        self.trace_matrix = np.ones((1, 1))

        # Flags
        self.active_buf_full_flag = False
        self.hashed_buffer_valid = False
        self.trace_valid = False
        self.use_ramulator_trace = False
        self.enable_layout_evaluation = False

    #
    def set_params(self, backing_buf_obj,
                   total_size_bytes=1, word_size=1, active_buf_frac=0.9,
                   hit_latency=1, backing_buf_bw=1, num_bank=1, num_port=2,
                   enable_layout_evaluation=False, use_ramulator_trace = False
                   ):
        """
        Method to set the ifmap/filter double buffered memory simulation parameters for
        housekeeping.
        """

        self.total_size_bytes = total_size_bytes
        self.word_size = word_size

        assert 0.5 <= active_buf_frac < 1, "Valid active buf frac [0.5,1)"
        self.active_buf_frac = round(active_buf_frac, 2)
        self.hit_latency = hit_latency

        # Calculate these based on the values provided
        self.total_size_elems = math.floor(self.total_size_bytes / self.word_size)
        self.active_buf_size = int(math.ceil(self.total_size_elems * self.active_buf_frac))
        self.prefetch_buf_size = self.total_size_elems - self.active_buf_size

        self.backing_buffer = backing_buf_obj
        self.req_gen_bandwidth = backing_buf_bw

        # Layout modeling
        self.num_bank = num_bank
        self.num_port = num_port # number of ports per bank
        self.bw_per_bank = self.req_gen_bandwidth // self.num_bank # bandwidth per bank
        self.enable_layout_evaluation = enable_layout_evaluation
        assert self.bw_per_bank * self.num_bank == self.req_gen_bandwidth, f"overall bandwidth must be divisible by total number of banks, number of banks = {self.num_bank}, bandwidth of each as {self.bw_per_bank}, total bandwidth = {self.req_gen_bandwidth}"

        # Ramulator trace
        self.use_ramulator_trace = use_ramulator_trace

    #
    def reset(self): # TODO: check if all resets are working propoerly
        """
        Method to reset the read buffer parameters.
        """
        # Buffer properties: User specified
        self.total_size_bytes = 128
        self.word_size = 1  # Bytes
        self.active_buf_frac = 0.9
        self.hit_latency = 1  # Cycles after which a request is served if already in the buffer

        # Buffer properties: Calculated
        self.total_size_elems = math.floor(self.total_size_bytes / self.word_size)
        self.active_buf_size = int(math.ceil(self.total_size_elems * 0.9))
        self.prefetch_buf_size = self.total_size_elems - self.active_buf_size

        # Backing interface properties
        self.backing_buffer = read_port()
        self.req_gen_bandwidth = 100  # words per cycle

        # Status of the buffer
        self.hashed_buffer = {}
        self.active_buffer_set_limits = []
        self.prefetch_buffer_set_limits = []

        # Variables to enable prefetching
        self.fetch_matrix = np.ones((1, 1))
        self.last_prefetch_cycle = -1
        self.next_line_prefetch_idx = 0
        self.next_col_prefetch_idx = 0

        # Access counts
        self.num_access = 0

        # Trace matrix
        self.trace_matrix = np.ones((1, 1))

        # Flags
        self.active_buf_full_flag = False
        self.hashed_buffer_valid = False
        self.trace_valid = False
        self.use_ramulator_trace = False

    #
    def set_fetch_matrix(self, fetch_matrix_np):
        """
        Method to set the fetch matrix responsible for prefetching from the DRAM.
        """
        # The operand matrix determines what to pre-fetch into both active and prefetch buffers
        # In 'user' mode, this will be set in the set_params

        num_elems = fetch_matrix_np.shape[0] * fetch_matrix_np.shape[1]
        num_lines = int(math.ceil(num_elems / self.req_gen_bandwidth))
        self.fetch_matrix = np.ones((num_lines, self.req_gen_bandwidth)) * -1

        # Put stuff into the fetch matrix
        # This is done to ensure that there is no shape mismatch
        # Not sure if this is the optimal way to do it or not
        for i in range(num_elems):
            src_row = math.floor(i / fetch_matrix_np.shape[1])
            src_col = math.floor(i % fetch_matrix_np.shape[1])

            dest_row = math.floor(i / self.req_gen_bandwidth)
            dest_col = math.floor(i % self.req_gen_bandwidth)

            self.fetch_matrix[dest_row][dest_col] = fetch_matrix_np[src_row][src_col]

        # Once the fetch matrices are set, populate the data structure for faster lookups and
        # servicing
        self.prepare_hashed_buffer()

    #
    def prepare_hashed_buffer(self):
        # layout modeling: hashed_buffer is being modified to serve as on-chip buffer.
        """
        Method to convert the fetch matrix into a hashed buffer for fast lookups.
        """
        elems_per_set = math.ceil(self.total_size_elems / 100)
        if self.enable_layout_evaluation:
            elems_per_set = self.req_gen_bandwidth
        
        prefetch_rows = self.fetch_matrix.shape[0]
        prefetch_cols = self.fetch_matrix.shape[1]

        line_id = 0
        elem_ctr = 0
        current_line = set()

        for r in range(prefetch_rows):
            for c in range(prefetch_cols):
                elem = self.fetch_matrix[r][c]

                if not elem == -1:
                    current_line.add(elem)
                    elem_ctr += 1

                if not elem_ctr < elems_per_set:    # ie > or =
                    self.hashed_buffer[line_id] = current_line
                    line_id += 1
                    elem_ctr = 0
                    current_line = set()        # new set

        self.hashed_buffer[line_id] = current_line

        max_num_active_buf_lines = int(math.ceil(self.active_buf_size / elems_per_set))
        max_num_prefetch_buf_lines = int(math.ceil(self.prefetch_buf_size / elems_per_set))
        num_lines = line_id + 1

        if num_lines > max_num_active_buf_lines:
            self.num_active_buf_lines = max_num_active_buf_lines
        else:
            self.num_active_buf_lines = num_lines

        remaining_lines = num_lines - self.num_active_buf_lines

        if remaining_lines > max_num_prefetch_buf_lines:
            self.num_prefetch_buf_lines = max_num_prefetch_buf_lines
        else:
            self.num_prefetch_buf_lines = remaining_lines

        self.num_lines = num_lines
        self.hashed_buffer_valid = True

    #
    def active_buffer_hit(self, addr):
        """
        Method to check if the address is hit or miss in the active read buffer.
        """
        assert self.active_buf_full_flag, 'Active buffer is not ready yet'

        start_id, end_id = self.active_buffer_set_limits
        if self.enable_layout_evaluation:
          if start_id < end_id:
              for line_id in range(start_id, end_id):
                  this_set = self.hashed_buffer[line_id]      # O(1) --> accessing hash
                  if addr in this_set:                        # Checking in a set(), O(1) lookup
                      return line_id, list(this_set).index(addr)

          else:
              for line_id in range(start_id, self.num_lines):
                  this_set = self.hashed_buffer[line_id]  # O(1) --> accessing hash
                  if addr in this_set:  # Checking in a set(), O(1) lookup
                      return line_id, list(this_set).index(addr)

              for line_id in range(end_id):
                  this_set = self.hashed_buffer[line_id]  # O(1) --> accessing hash
                  if addr in this_set:  # Checking in a set(), O(1) lookup
                      return line_id, list(this_set).index(addr)
          # Fixing for ISSUE #14
          # return True
          return -1, -1
        else:
          if start_id < end_id:
              for line_id in range(start_id, end_id):
                  this_set = self.hashed_buffer[line_id]      # O(1) --> accessing hash
                  if addr in this_set:                        # Checking in a set(), O(1) lookup
                      return True

          else:
              for line_id in range(start_id, self.num_lines):
                  this_set = self.hashed_buffer[line_id]  # O(1) --> accessing hash
                  if addr in this_set:  # Checking in a set(), O(1) lookup
                      return True

              for line_id in range(end_id):
                  this_set = self.hashed_buffer[line_id]  # O(1) --> accessing hash
                  if addr in this_set:  # Checking in a set(), O(1) lookup
                      return True
          # Fixing for ISSUE #14
          # return True
          return False

    #
    def service_reads(self,
                      incoming_requests_arr_np,   # 2D array with the requests
                      incoming_cycles_arr):       # 1D vector with the cycles at which req arrived
        """
        Method to service read requests coming from systolic array. Logic: Always check if an addr
        is in active buffer. If hit, return with hit latency Else, make the contents of prefetch
        buffer as active and then check Continue making new prefetches until there is a hit.
        """
        # Service the incoming read requests
        # returns a cycles array corresponding to the requests buffer
        # Logic: Always check if an addr is in active buffer.
        #        If hit, return with hit latency
        #        Else, make the contents of prefetch buffer as active and then check
        #              finish till an ongoing prefetch is done before reassiging prefetch buffer
        dram_stall_cycles = 0
        if not self.active_buf_full_flag:
            start_cycle = incoming_cycles_arr[0][0]
            # Needs to use the entire operand matrix
            # keeping in mind the tile order and everything
            dram_stall_cycles = self.prefetch_active_buffer(start_cycle=start_cycle)

        out_cycles_arr = []
        offset = self.hit_latency
        if self.enable_layout_evaluation:
          for i in tqdm(range(incoming_requests_arr_np.shape[0]), disable=True):
              cycle = incoming_cycles_arr[i]
              # Fixing for ISSUE #14
              # request_line = set(incoming_requests_arr_np[i]) #shaves off a few seconds
              request_line = incoming_requests_arr_np[i]

              concurrent_line_addr = [[] for _ in range(self.num_bank)] # bank conflict modeling
              for addr in request_line:
                  if addr == -1:
                      continue

                  # if addr not in self.active_buffer_contents: #this is super slow!!!
                  # Fixing for ISSUE #14
                  # if not self.active_buffer_hit(addr):  # --> While loop ensures multiple prefetches if needed
                  line_addr, column_addr = self.active_buffer_hit(addr)
                  while line_addr == -1:
                      self.new_prefetch()
                      potential_stall_cycles = self.last_prefetch_cycle - (cycle + offset)

                      if potential_stall_cycles > 0:
                          offset += potential_stall_cycles
                      line_addr, column_addr = self.active_buffer_hit(addr)
                  
                  # Layout Modeling 1 -- data mapping to multiple bank 
                  # The 2D array is interleaved mapped to multiple banks. 
                  # e.g. (interleave mapping) addr 0 -> bank 0; addr 1 -> bank 1; addr 2 -> bank 2 ...
                  # instead of (contiguous mapping) addr 0,...,lines in a bank - 1 -> bank 0.
                  # because data access in compiled layout turns to be contiguious, which accesses the continuous addresses.
                  # In (contiguous mapping), such multi-bank mapping would result in more bank conflict slowdown.
                  bank_id = column_addr // self.bw_per_bank
                  assert bank_id < self.num_bank, f"bank id = {bank_id} for column_addr = {column_addr} needs to be smaller than total number of bank = {self.num_bank}"
                  if line_addr not in concurrent_line_addr[bank_id]:
                      concurrent_line_addr[bank_id].append(line_addr)
              max_line_request_among_all_banks = 0
              for bank_id in range(self.num_bank):
                  max_line_request_among_all_banks = max(len(concurrent_line_addr[bank_id]), max_line_request_among_all_banks)
              offset += math.ceil(max_line_request_among_all_banks/self.num_port) - 1
              if self.use_ramulator_trace == True:
                  out_cycles = cycle + offset + dram_stall_cycles
              else:
                  out_cycles = cycle + offset
              out_cycles_arr.append(out_cycles)

          out_cycles_arr_np = np.asarray(out_cycles_arr).reshape((len(out_cycles_arr), 1))

          return out_cycles_arr_np
        
        else:
          for i in tqdm(range(incoming_requests_arr_np.shape[0]), disable=True):
              cycle = incoming_cycles_arr[i]
              # Fixing for ISSUE #14
              # request_line = set(incoming_requests_arr_np[i]) #shaves off a few seconds
              request_line = incoming_requests_arr_np[i]

              for addr in request_line:
                  if addr == -1:
                      continue

                  # if addr not in self.active_buffer_contents: #this is super slow!!!
                  # Fixing for ISSUE #14
                  # if not self.active_buffer_hit(addr):  # --> While loop ensures multiple prefetches if needed
                  while not self.active_buffer_hit(addr):
                      self.new_prefetch()
                      potential_stall_cycles = self.last_prefetch_cycle - (cycle + offset)
                      offset += potential_stall_cycles        # Offset increments if there were potential stalls
                      if potential_stall_cycles > 0:
                          offset += potential_stall_cycles
                    
              if self.use_ramulator_trace == True:
                  out_cycles = cycle + offset + dram_stall_cycles
              else:
                  out_cycles = cycle + offset
              out_cycles_arr.append(out_cycles)

          out_cycles_arr_np = np.asarray(out_cycles_arr).reshape((len(out_cycles_arr), 1))

          return out_cycles_arr_np

    #
    def prefetch_active_buffer(self, start_cycle):
        """
        Method to prefetch the active read buffer before servicing individual memory requests.
        """
        # Depending on size of the active buffer, calculate the number of lines from op mat to fetch
        # Also, calculate the cycles arr for requests

        # 1. Preparing the requests:
        num_lines = math.ceil(self.active_buf_size / self.req_gen_bandwidth)
        if not num_lines < self.fetch_matrix.shape[0]:
            num_lines = self.fetch_matrix.shape[0]

        requested_data_size = num_lines * self.req_gen_bandwidth
        self.num_access += requested_data_size

        start_idx = 0
        end_idx = num_lines

        prefetch_requests = self.fetch_matrix[start_idx:end_idx, :]

        # 1.1 See if extra requests are made, if so nullify them
        self.next_col_prefetch_idx = 0
        if requested_data_size > self.active_buf_size:
            valid_cols = int(self.active_buf_size % self.req_gen_bandwidth)
            row = end_idx - 1
            self.next_col_prefetch_idx = valid_cols
            for col in range(valid_cols, self.req_gen_bandwidth):
                prefetch_requests[row][col] = -1

        # TODO: Tally and check if this agrees with the contents of the hashed buffer

        # 2. Preparing the cycles array
        #    The start_cycle variable ensures that all the requests have been made before any
        #    incoming reads came
        cycles_arr = np.zeros((num_lines, 1))
        for i in range(cycles_arr.shape[0]):
            cycles_arr[i][0] = \
                -1 * (num_lines - start_cycle - (i - self.backing_buffer.get_latency()))

        # 3. Send the request and get the response cycles count
        response_cycles_arr = \
            self.backing_buffer.service_reads(incoming_cycles_arr=cycles_arr,
                                              incoming_requests_arr_np=prefetch_requests)

        # 4. Update the variables
        #self.last_prefetch_cycle = int(response_cycles_arr[-1][0])
        self.last_prefetch_cycle = int(max(response_cycles_arr))

        # Update the trace matrix
        self.trace_matrix = np.column_stack((response_cycles_arr, prefetch_requests))
        self.trace_valid = True

        # Set active buffer contents
        active_buf_start_line_id = 0
        active_buf_end_line_id = self.num_active_buf_lines
        self.active_buffer_set_limits = [active_buf_start_line_id, active_buf_end_line_id]

        prefetch_buf_start_line_id = active_buf_end_line_id
        prefetch_buf_end_line_id = prefetch_buf_start_line_id + self.num_prefetch_buf_lines
        self.prefetch_buffer_set_limits = [prefetch_buf_start_line_id, prefetch_buf_end_line_id]

        self.active_buf_full_flag = True

        # Set the line to be prefetched next
        # The module operator is to ensure that the indices wrap around
        if requested_data_size > self.active_buf_size:
            # Some elements in the current idx is left out in this case
            self.next_line_prefetch_idx = num_lines % self.fetch_matrix.shape[0]
        else:
            self.next_line_prefetch_idx = (num_lines + 1) % self.fetch_matrix.shape[0]

        return (self.last_prefetch_cycle - cycles_arr[-1][0] - 1)   
    #
    def new_prefetch(self):
        """
        Method to do a new prefetch. In a new prefetch, some portion of the original data needs to
        be deleted to accomodate the prefetched data In this case we overwrite some data in the
        active buffer with the prefetched data and then create a new prefetch request.
        """
        # In a new prefetch, some portion of the original data needs to be deleted to accomodate the
        # prefetched data.
        # In this case we overwrite some data in the active buffer with the prefetched data
        # And then create a new prefetch request
        # Also return when the prefetched data was made available

        # 1. Rewrite the active buffer
        assert self.active_buf_full_flag, 'Active buffer is empty'
        active_start, active_end = self.active_buffer_set_limits

        active_start = int((active_start + self.num_prefetch_buf_lines) % self.num_lines)
        active_end = int((active_start + self.num_active_buf_lines) % self.num_lines)
        prefetch_start = active_end
        prefetch_end = int((prefetch_start + self.num_prefetch_buf_lines) % self.num_lines)

        self.active_buffer_set_limits = [active_start, active_end]
        self.prefetch_buffer_set_limits = [prefetch_start, prefetch_end]

        # 2. Create the request
        start_idx = self.next_line_prefetch_idx
        num_lines = math.ceil(self.prefetch_buf_size / self.req_gen_bandwidth)
        end_idx = start_idx + num_lines
        requested_data_size = num_lines * self.req_gen_bandwidth
        self.num_access += requested_data_size

        # In case we need to circle back
        if end_idx > self.fetch_matrix.shape[0]:
            last_idx = self.fetch_matrix.shape[0]
            prefetch_requests = self.fetch_matrix[start_idx:,:]

            new_end_idx = min(end_idx - last_idx, start_idx) # In case the entire array is engulfed
            prefetch_requests = \
                np.concatenate((prefetch_requests, self.fetch_matrix[:new_end_idx,:]))
        else:
            prefetch_requests = self.fetch_matrix[start_idx:end_idx, :]

        # Modify the prefetch request to drop unwanted addresses
        # a. Chomp the elements in the first line included in previous fetches
        for i in range(0, self.next_col_prefetch_idx):
            prefetch_requests[0][i] = -1

        # b. Chomp the excess elements in the last line
        if requested_data_size > self.active_buf_size:
            valid_cols = int(self.active_buf_size % self.req_gen_bandwidth)
            row = prefetch_requests.shape[0] - 1
            for col in range(valid_cols, self.req_gen_bandwidth):
                prefetch_requests[row][col] = -1

        # 3. Create the request cycles
        cycles_arr = np.zeros((num_lines, 1))
        for i in range(cycles_arr.shape[0]):
            # Fixing ISSUE #14
            # cycles_arr[i][0] = self.last_prefetch_cycle + i
            cycles_arr[i][0] = self.last_prefetch_cycle + i + 1

        # 4. Send the request
        response_cycles_arr = \
            self.backing_buffer.service_reads(incoming_cycles_arr=cycles_arr,
                                              incoming_requests_arr_np=prefetch_requests)

        # 5. Update the variables
        self.last_prefetch_cycle = np.amax(response_cycles_arr)
        #assert response_cycles_arr.shape == cycles_arr.shape, \
        #       'The request and response cycles dims do not match'

        this_prefetch_trace = np.column_stack((response_cycles_arr, prefetch_requests))
        self.trace_matrix = np.concatenate((self.trace_matrix, this_prefetch_trace), axis=0)

        # Set the line to be prefetched next
        if requested_data_size > self.active_buf_size:
            self.next_line_prefetch_idx = num_lines % self.fetch_matrix.shape[0]
        else:
            self.next_line_prefetch_idx = (num_lines + 1) % self.fetch_matrix.shape[0]

        # This does not need to return anything

    #
    def get_trace_matrix(self):
        """
        Method to get the read buffer trace matrix. It contains addresses requsted by the systolic
        array and the cycles (first column) at which the requests are made.
        """
        if not self.trace_valid:
            print('No trace has been generated yet')
            return

        return self.trace_matrix

    #
    def get_hit_latency(self):
        """
        Method to get hit latency of the read buffer.
        """
        return self.hit_latency

    #
    def get_latency(self):
        """
        Method to get hit latency of the read buffer.
        """
        return self.hit_latency

    #
    def get_num_accesses(self):
        """
        Method to get number of accesses of the read buffer if trace_valid flag is set.
        """
        assert self.trace_valid, 'Traces not ready yet'
        return self.num_access

    #
    def get_external_access_start_stop_cycles(self):
        """
        Method to get start and stop cycles of the read buffer if trace_valid flag is set.
        """
        assert self.trace_valid, 'Traces not ready yet'
        start_cycle = np.amin(self.trace_matrix[:,0])
        end_cycle = np.amax(self.trace_matrix[:,0])

        return start_cycle, end_cycle

    #
    def print_trace(self, filename):
        """
        Method to write the read buffer trace matrix to a file.
        """
        if not self.trace_valid:
            print('No trace has been generated yet')
            return

        np.savetxt(filename, self.trace_matrix, fmt='%s', delimiter=",")


"""
This file contains the 'compression' class that defines the various compression formats such as
Compressed Sparse Row, Compressed Sparse Column and Blocked ELLPACK. The class provides methods to
obtain the Filter and Metadata storages along with original storage for all the compression formats
mentioned.
"""

import math
import numpy as np

class compression:
    """
    This class deals with calculating the filter storage for with/without sparsity cases.
    """
    #
    def compress_to_csr(self, matrix):
        """
        Method to compress a given matrix using the Compressed Sparse Row (CSR) format.
        """
        rows, cols = matrix.shape
        data = []
        col_id = []
        row_ptr = [0]

        for i in range(rows):
            # Find non-zero elements in current row
            non_zero_indices = np.nonzero(matrix[i])[0]
            data.extend(matrix[i, non_zero_indices])
            col_id.extend(non_zero_indices)
            row_ptr.append(len(data))  # Update starting index for next row

        original_storage = rows * cols
        new_storage = (2 * len(col_id)) + len(row_ptr)
        metadata_storage = len(col_id) + len(row_ptr)

        return np.array(data), np.array(col_id), np.array(row_ptr), original_storage, new_storage, \
               metadata_storage

    #
    def compress_to_csc(self, matrix):
        """
        Method to compress a given matrix using the Compressed Sparse Column (CSC) format.
        """
        rows, cols = matrix.shape
        data = []
        row_id = []
        col_ptr = [0] * (cols + 1)

        for j in range(cols):
            # Find non-zero elements in current column
            non_zero_indices = np.nonzero(matrix[:, j])[0]
            data.extend(matrix[non_zero_indices, j])
            row_id.extend(non_zero_indices)
            col_ptr[j + 1] = len(data)  # Update starting index for next column

        original_storage = rows * cols
        new_storage = (2 * len(row_id)) + len(col_ptr)
        metadata_storage = len(row_id) + len(col_ptr)

        return np.array(data), np.array(row_id), np.array(col_ptr), original_storage, new_storage, \
               metadata_storage

    #
    def compress_to_ellpack_block(self, matrix, filter_op_mat, sparsity_ratio_M):
        """
        Method to compress a given matrix using the Blocked ELLPACK format.
        """
        # Each entry of the original matrix requires 1 word size.
        # Each entry in filter_op_mat will require log2(M) bits of metadata.
        original_rows, original_cols = matrix.shape
        new_rows, new_cols = filter_op_mat.shape

        # Units: Words (considering 1 word = 4 bytes = 32 bits)
        original_storage = original_rows * original_cols
        metadata_storage = ((new_rows * new_cols) * math.ceil(math.log2(sparsity_ratio_M))) / 32
        new_storage = (new_rows * new_cols) + metadata_storage

        return original_storage, new_storage, metadata_storage

    #
    def get_csr_storage(self, matrix):
        """
        Method to get the sizes of original filter, compressed dense filter matrix and its metadata
        when CSR is used as the comprtession format.
        """
        data, col_id, row_ptr, original_storage, new_storage, metadata_storage = \
            self.compress_to_csr(matrix)
        return original_storage, new_storage, metadata_storage

    #
    def get_csc_storage(self, matrix):
        """
        Method to get the sizes of original filter, compressed dense filter matrix and its metadata
        when CSC is used as the comprtession format.
        """
        data, row_id, col_ptr, original_storage, new_storage, metadata_storage = \
            self.compress_to_csc(matrix)
        return original_storage, new_storage, metadata_storage

    #
    def get_ellpack_block_storage(self, matrix, filter_op_mat, sparsity_ratio_M):
        """
        Method to get the sizes of original filter, compressed dense filter matrix and its metadata
        when Blocked ELLPACK is used as the comprtession format.
        """
        original_storage, new_storage, metadata_storage = \
            self.compress_to_ellpack_block(matrix, filter_op_mat, sparsity_ratio_M)
        return original_storage, new_storage, metadata_storage
